# -*- coding: utf-8 -*-
import os
import re
import sys
import subprocess
from airtestProject.airtest.utils.compat import decode_path

THISPATH = decode_path(os.path.dirname(os.path.realpath(__file__)))
STATICPATH = os.path.join(THISPATH, "static")
DEFAULT_ADB_PATH = {
    "Windows": os.path.join(STATICPATH, "adb", "windows", "adb.exe"),
    "Darwin": os.path.join(STATICPATH, "adb", "mac", "adb"),
    "Linux": os.path.join(STATICPATH, "adb", "linux", "adb"),
    "Linux-x86_64": os.path.join(STATICPATH, "adb", "linux", "adb"),
    "Linux-armv7l": os.path.join(STATICPATH, "adb", "linux_arm", "adb"),
}
DEFAULT_ADB_SERVER = ('127.0.0.1', 5037)
SDK_VERISON_ANDROID7 = 24
# Android 10 SDK version
SDK_VERISON_ANDROID10 = 29
DEBUG = True
STFLIB = os.path.join(STATICPATH, "stf_libs")
ROTATIONWATCHER_APK = os.path.join(STATICPATH, "apks", "RotationWatcher.apk")
ROTATIONWATCHER_PACKAGE = "jp.co.cyberagent.stf.rotationwatcher"
YOSEMITE_APK = os.path.join(STATICPATH, "apks", "Yosemite.apk")
YOSEMITE_PACKAGE = 'com.netease.nie.yosemite'
YOSEMITE_IME_SERVICE = 'com.netease.nie.yosemite/.ime.ImeService'
MAXTOUCH_JAR = os.path.join(STATICPATH, "apks", "maxpresent.jar")
ROTATIONWATCHER_JAR = os.path.join(STATICPATH, "apks", "rotationwatcher.jar")
IP_PATTERN = re.compile(r'(\d+\.){3}\d+')


class CAP_METHOD(object):
    MINICAP = "MINICAP"
    ADBCAP = "ADBCAP"
    JAVACAP = "JAVACAP"


class TOUCH_METHOD(object):
    MINITOUCH = "MINITOUCH"
    MAXTOUCH = "MAXTOUCH"
    ADBTOUCH = "ADBTOUCH"


class IME_METHOD(object):
    ADBIME = "ADBIME"
    YOSEMITEIME = "YOSEMITEIME"


class ORI_METHOD(object):
    ADB = "ADBORI"
    MINICAP = "MINICAPORI"



# coding=utf-8

from ..interfaces import IClient
from .safetcp import Client, socket
from .protocol import SimpleProtocolFilter


DEFAULT_ADDR = ("0.0.0.0", 5001)


class TcpClient(IClient):
    """docstring for TcpClient"""
    def __init__(self, addr=DEFAULT_ADDR):
        super(TcpClient, self).__init__()
        self.addr = addr
        self.prot = SimpleProtocolFilter()
        self.c = None

    def __str__(self):
        return 'tcp://{}:{}'.format(*self.addr)
    __repr__ = __str__

    def connect(self):
        if not self.c:
            self.c = Client(self.addr,
                            on_connect=self.on_connect,
                            on_close=self.on_close)
        self.c.connect()

    def send(self, msg):
        msg_bytes = self.prot.pack(msg)
        self.c.send(msg_bytes)

    def recv(self):
        try:
            msg_bytes = self.c.recv()
        except socket.timeout:
            # print("socket recv timeout")
            msg_bytes = b""
        return self.prot.input(msg_bytes)

    def close(self):
        self.c.close()
        self.c = None



# -*- coding: utf-8 -*-
import win32gui
import win32api
import win32ui
import win32con
from airtestProject.airtest.aircv.utils import Image, pil_2_cv2


SM_XVIRTUALSCREEN = 76
SM_YVIRTUALSCREEN = 77
SM_CXVIRTUALSCREEN = 78
SM_CYVIRTUALSCREEN = 79


def screenshot(filename, hwnd=None):
    """
    Take the screenshot of Windows app

    Args:
        filename: file name where to store the screenshot
        hwnd:

    Returns:
        bitmap screenshot file

    """
    # import ctypes
    # user32 = ctypes.windll.user32
    # user32.SetProcessDPIAware()

    if hwnd is None:
        """all screens"""
        hwnd = win32gui.GetDesktopWindow()
        # get complete virtual screen including all monitors
        w = win32api.GetSystemMetrics(SM_CXVIRTUALSCREEN)
        h = win32api.GetSystemMetrics(SM_CYVIRTUALSCREEN)
        x = win32api.GetSystemMetrics(SM_XVIRTUALSCREEN)
        y = win32api.GetSystemMetrics(SM_YVIRTUALSCREEN)
    else:
        """window"""
        rect = win32gui.GetWindowRect(hwnd)
        w = abs(rect[2] - rect[0])
        h = abs(rect[3] - rect[1])
        x, y = 0, 0
    hwndDC = win32gui.GetWindowDC(hwnd)
    mfcDC = win32ui.CreateDCFromHandle(hwndDC)
    saveDC = mfcDC.CreateCompatibleDC()
    saveBitMap = win32ui.CreateBitmap()
    saveBitMap.CreateCompatibleBitmap(mfcDC, w, h)
    saveDC.SelectObject(saveBitMap)
    saveDC.BitBlt((0, 0), (w, h), mfcDC, (x, y), win32con.SRCCOPY)
    # saveBitMap.SaveBitmapFile(saveDC, filename)
    bmpinfo = saveBitMap.GetInfo()
    bmpstr = saveBitMap.GetBitmapBits(True)
    pil_image = Image.frombuffer(
        'RGB',
        (bmpinfo['bmWidth'], bmpinfo['bmHeight']),
        bmpstr, 'raw', 'BGRX', 0, 1)
    cv2_image = pil_2_cv2(pil_image)

    mfcDC.DeleteDC()
    saveDC.DeleteDC()
    win32gui.ReleaseDC(hwnd, hwndDC)
    win32gui.DeleteObject(saveBitMap.GetHandle())
    return cv2_image



# _*_ coding:UTF-8 _*_
import socket
import errno


class SafeSocket(object):
    """safe and exact recv & send"""
    def __init__(self, sock=None):
        if sock is None:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        else:
            self.sock = sock
        self.buf = b""

    def __enter__(self):
        try:
            return self.sock.__enter__()
        except AttributeError:
            return self.sock

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            return self.sock.__exit__(exc_type, exc_val, exc_tb)
        except AttributeError:
            self.sock.close()

    # PEP 3113 -- Removal of Tuple Parameter Unpacking
    # https://www.python.org/dev/peps/pep-3113/
    def connect(self, tuple_hp):
        host, port = tuple_hp
        self.sock.connect((host, port))

    def send(self, msg):
        totalsent = 0
        while totalsent < len(msg):
            sent = self.sock.send(msg[totalsent:])
            if sent == 0:
                raise socket.error("socket connection broken")
            totalsent += sent

    def recv(self, size):
        while len(self.buf) < size:
            trunk = self.sock.recv(min(size-len(self.buf), 4096))
            if trunk == b"":
                raise socket.error("socket connection broken")
            self.buf += trunk
        ret, self.buf = self.buf[:size], self.buf[size:]
        return ret

    def recv_with_timeout(self, size, timeout=2):
        self.sock.settimeout(timeout)
        try:
            ret = self.recv(size)
        except socket.timeout:
            ret = None
        finally:
            self.sock.settimeout(None)
        return ret

    def recv_nonblocking(self, size):
        self.sock.settimeout(0)
        try:
            ret = self.recv(size)
        except(socket.error) as e:
            #10035 no data when nonblocking
            if e.args[0] == 10035: #errno.EWOULDBLOCK: 尼玛errno似乎不一致
                ret = None
            #10053 connection abort by client
            #10054 connection reset by peer
            elif e.args[0] in [10053, 10054]: #errno.ECONNABORTED:
                raise
            else:
                raise
        return ret

    def close(self):
        if hasattr(self.sock, "_closed") and not self.sock._closed:
            try:
                self.sock.shutdown(socket.SHUT_RDWR)
            except OSError as e:
                if e.errno != errno.ENOTCONN:  # 'Socket is not connected'
                    raise
            self.sock.close()
        else:
            self.sock.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Declaration:
    Define all BaseError Classes used in aircv.
"""


class BaseError(Exception):
    """Base class for exceptions in this module."""

    def __init__(self, message=""):
        self.message = message

    def __repr__(self):
        return repr(self.message)


class FileNotExistError(BaseError):
    """Image does not exist."""


class TemplateInputError(BaseError):
    """Resolution input is not right."""


class NoSIFTModuleError(BaseError):
    """Resolution input is not right."""


class NoSiftMatchPointError(BaseError):
    """Exception raised for errors 0 sift points found in the input images."""


class SiftResultCheckError(BaseError):
    """Exception raised for errors 0 sift points found in the input images."""


class HomographyError(BaseError):
    """In homography, find no mask, should kill points which is duplicate."""


class NoModuleError(BaseError):
    """Resolution input is not right."""


class NoMatchPointError(BaseError):
    """Exception raised for errors 0 keypoint found in the input images."""


class MatchResultCheckError(BaseError):
    """Exception raised for errors 0 keypoint found in the input images."""



# coding=utf-8

from airtestProject.airtest.core.api import connect_device


def UnityEditorWindow():
    dev = connect_device("Windows:///?class_name=UnityContainerWndClass&title_re=.*Unity.*")
    game_window = dev.app.top_window().child_window(title="UnityEditor.GameView")
    dev._top_window = game_window.wrapper_object()
    dev.focus_rect = (0, 40, 0, 0)
    return dev



import json
import os

import shutil


from airtestProject.airtest.report.report import LogToHtml


try:
    from jinja2 import evalcontextfilter as pass_eval_context  # jinja2<3.1
except:
    from jinja2 import pass_eval_context  # jinja2>=3.1

from airtestProject.airtest.core.settings import Settings as ST
from airtestProject.airtest.aircv import imread, get_resolution, FileNotExistError
from airtestProject.airtest.aircv.utils import compress_image
from airtestProject.airtest.utils.compat import decode_path, script_dir_name
from airtestProject.airtest.cli.info import get_script_info
from airtestProject.airtest.utils.logger import get_logger


LOGGING = get_logger(__name__)
DEFAULT_LOG_DIR = "log"
DEFAULT_LOG_FILE = "log.txt"
HTML_TPL = "log_template.html"
HTML_FILE = "log.html"
STATIC_DIR = os.path.dirname(__file__)


class ReportUtil(LogToHtml):
    def __init__(self, script_root, log_root="", static_root="", export_dir=None, script_name="", logfile=None, lang="en", plugins=None):
        self.log = []
        self.devices = {}
        self.script_root = script_root
        self.script_name = script_name
        if not self.script_name or os.path.isfile(self.script_root):
            self.script_root, self.script_name = script_dir_name(self.script_root)
        self.log_root = log_root or ST.LOG_DIR or os.path.join(".", DEFAULT_LOG_DIR)
        self.static_root = static_root or STATIC_DIR
        self.test_result = True
        self.run_start = None
        self.run_end = None
        self.export_dir = export_dir
        self.logfile = logfile or getattr(ST, "LOG_FILE", DEFAULT_LOG_FILE)
        self.lang = lang
        self.init_plugin_modules(plugins)
        super().__init__(script_root, log_root, static_root, export_dir, script_name, logfile,
                         lang, plugins)

    def _make_export_dir(self):
        """mkdir & copy /staticfiles/screenshots"""
        # let dirname = <script name>.log
        dirname = self.script_name.replace(os.path.splitext(self.script_name)[1], ".log")
        # mkdir
        dirpath = os.path.join(self.export_dir, dirname)
        if os.path.isdir(dirpath):
            shutil.rmtree(dirpath, ignore_errors=True)

        # copy script
        def ignore_export_dir(dirname, filenames):
            # 忽略当前导出的目录，防止递归导出
            if os.path.commonprefix([dirpath, dirname]) == dirpath:
                return filenames
            return []
        # self.copy_tree(self.script_root, dirpath, ignore=ignore_export_dir)
        # copy log
        logpath = os.path.join(dirpath, DEFAULT_LOG_DIR)
        if os.path.normpath(logpath) != os.path.normpath(self.log_root):
            if os.path.isdir(logpath):
                shutil.rmtree(logpath, ignore_errors=True)
            self.copy_tree(self.log_root, logpath, ignore=shutil.ignore_patterns(dirname))
        # if self.static_root is not a http server address, copy static files from local directory
        if not self.static_root.startswith("http"):
            for subdir in ["css", "fonts", "image", "js"]:
                self.copy_tree(os.path.join(self.static_root, subdir), os.path.join(dirpath, "static", subdir))

        return dirpath, logpath

    def _translate_code(self, step):
        if step["tag"] != "function":
            return None
        step_data = step["data"]
        args = []
        code = {
            "name": step_data["name"],
            "args": args,
        }
        for key, value in step_data["call_args"].items():
            args.append({
                "key": key,
                "value": value,
            })
        for k, arg in enumerate(args):
            value = arg["value"]
            if isinstance(value, dict) and value.get("__class__") == "OcrTemplate":
                arg["value"] = value['filename']
            if isinstance(value, dict) and (value.get("__class__") == "Template" or value.get("__class__")
                                            == "myTemplate"):
                if self.export_dir:  # all relative path
                    image_path = value['filename']
                    if not os.path.isfile(os.path.join(self.script_root, image_path)) and value['_filepath']:
                        # copy image used by using statement
                        shutil.copyfile(value['_filepath'], os.path.join(self.script_root, value['filename']))
                else:
                    image_path = os.path.abspath(value['_filepath'] or value['filename'])
                arg["image"] = image_path
                try:
                    if not value['_filepath'] and not os.path.exists(value['filename']):
                        crop_img = imread(os.path.join(self.script_root, value['filename']))
                    else:
                        crop_img = imread(value['_filepath'] or value['filename'])
                except FileNotExistError:
                    # 在某些情况下会报图片不存在的错误（偶现），但不应该影响主流程
                    if os.path.exists(image_path):
                        arg["resolution"] = get_resolution(imread(image_path))
                    else:
                        arg["resolution"] = (0, 0)
                else:
                    arg["resolution"] = get_resolution(crop_img)
        return code

    def report_data(self, output_file=None, record_list=None):
        """
        Generate data for the report page

        :param output_file: The file name or full path of the output file, default HTML_FILE
        :param record_list: List of screen recording files
        :return:
        """
        self._load()
        steps = self._analyse()

        script_path = os.path.join(self.script_root, self.script_name)
        info = json.loads(get_script_info(script_path))
        info['devices'] = self.devices

        if record_list:
            records = [os.path.join(DEFAULT_LOG_DIR, f) if self.export_dir
                       else os.path.abspath(os.path.join(self.log_root, f)) for f in record_list]
        else:
            records = []

        if not self.static_root.endswith(os.path.sep):
            self.static_root = self.static_root.replace("\\", "/")
            self.static_root += "/"

        if not output_file:
            output_file = HTML_FILE

        data = {}
        data['steps'] = steps
        data['name'] = self.script_root
        data['scale'] = self.scale
        data['test_result'] = self.test_result
        data['run_end'] = self.run_end
        data['run_start'] = self.run_start
        data['static_root'] = self.static_root
        data['lang'] = self.lang
        data['records'] = records
        data['info'] = info
        data['log'] = self.get_relative_log(output_file)
        data['console'] = self.get_console(output_file)
        # 如果带有<>符号，容易被highlight.js认为是特殊语法，有可能导致页面显示异常，尝试替换成不常用的{}
        info = json.dumps(data).replace("<", "{").replace(">", "}")
        data['data'] = info
        return data

    def report(self, template_name=HTML_TPL, output_file=HTML_FILE, record_list=None):
        """
        Generate the reports page, you can add custom data and overload it if needed

        :param template_name: default is HTML_TPL
        :param output_file: The file name or full path of the output file, default HTML_FILE
        :param record_list: List of screen recording files
        :return:
        """
        if not self.script_name:
            path, self.script_name = script_dir_name(self.script_root)

        if self.export_dir:
            script_root, self.log_root = self._make_export_dir()
            # output_file可传入文件名，或绝对路径
            output_file = output_file if output_file and os.path.isabs(output_file) \
                else os.path.join(script_root, output_file or HTML_FILE)
            if not self.static_root.startswith("http"):
                self.static_root = "static/"

        if not record_list:
            record_list = [f for f in os.listdir(self.log_root) if f.endswith(".mp4")]
        data = self.report_data(output_file=output_file, record_list=record_list)
        return self._render(template_name, output_file, **data)



# -*- coding: utf-8 -*-
"""
This module contains the Airtest Core APIs.
"""
import os
import time

from airtestProject.airtest.core.cv import Template, loop_find, try_log_screen
from airtestProject.airtest.core.error import TargetNotFoundError
from airtestProject.airtest.core.settings import Settings as ST
from airtestProject.airtest.utils.compat import script_log_dir
from airtestProject.airtest.utils.snippet import parse_device_uri
from airtestProject.airtest.core.helper import (G, delay_after_operation, import_device_cls,
                                 logwrap, set_logdir, using, log)
# Assertions
from airtestProject.airtest.core.assertions import (assert_exists, assert_not_exists, assert_equal, assert_not_equal,  # noqa
                                        assert_true, assert_false, assert_is, assert_is_not,
                                        assert_is_none, assert_is_not_none, assert_in, assert_not_in,
                                        assert_is_instance, assert_not_is_instance
                                     )


"""
Device Setup APIs
"""


def init_device(platform="Android", uuid=None, **kwargs):
    """
    Initialize device if not yet, and set as current device.

    :param platform: Android, IOS or Windows
    :param uuid: uuid for target device, e.g. serialno for Android, handle for Windows, uuid for iOS
    :param kwargs: Optional platform specific keyword args, e.g. `cap_method=JAVACAP` for Android
    :return: device instance
    :Example:

        >>> init_device(platform="Android",uuid="SJE5T17B17", cap_method="JAVACAP")
        >>> init_device(platform="Windows",uuid="123456")
    """
    cls = import_device_cls(platform)
    dev = cls(uuid, **kwargs)
    # Add device instance in G and set as current device.
    G.add_device(dev)
    return dev


@logwrap
def connect_device(uri):
    """
    Initialize device with uri, and set as current device.

    :param uri: an URI where to connect to device, e.g. `android://adbhost:adbport/serialno?param=value&param2=value2`
    :return: device instance
    :Example:

        >>> connect_device("Android:///")  # local adb device using default params
        >>> # local device with serial number SJE5T17B17 and custom params
        >>> connect_device("Android:///SJE5T17B17?cap_method=javacap&touch_method=adb")
        >>> # remote device using custom params Android://adbhost:adbport/serialno
        >>> connect_device("Android://127.0.0.1:5037/10.254.60.1:5555")
        >>> connect_device("Android://127.0.0.1:5037/10.234.60.1:5555?name=serialnumber")  # add serialno to params
        >>> connect_device("Windows:///")  # connect to the desktop
        >>> connect_device("Windows:///123456")  # Connect to the window with handle 123456
        >>> connect_device("windows:///?title_re='.*explorer.*'")  # Connect to the window that name include "explorer"
        >>> connect_device("Windows:///123456?foreground=False")  # Connect to the window without setting it foreground
        >>> connect_device("iOS:///127.0.0.1:8100")  # iOS device
        >>> connect_device("iOS:///http://localhost:8100/?mjpeg_port=9100")  # iOS with mjpeg port
        >>> connect_device("iOS:///http://localhost:8100/?mjpeg_port=9100&&udid=00008020-001270842E88002E")  # iOS with mjpeg port and udid
        >>> connect_device("iOS:///http://localhost:8100/?mjpeg_port=9100&&uuid=00008020-001270842E88002E")  # udid/uuid/serialno are all ok

    """
    platform, uuid, params = parse_device_uri(uri)
    dev = init_device(platform, uuid, **params)
    return dev


def device():
    """
    Return the current active device.

    :return: current device instance
    :Example:
        >>> dev = device()
        >>> dev.touch((100, 100))
    """
    return G.DEVICE


def set_current(idx):
    """
    Set current active device.

    :param idx: uuid or index of initialized device instance
    :raise IndexError: raised when device idx is not found
    :return: None
    :platforms: Android, iOS, Windows
    :Example:
        >>> # switch to the first phone currently connected
        >>> set_current(0)
        >>> # switch to the phone with serial number serialno1
        >>> set_current("serialno1")

    """
    # 创建一个字典。以dev的uuid属性为字典的key
    dev_dict = {dev.uuid: dev for dev in G.DEVICE_LIST}
    if idx in dev_dict:
        current_dev = dev_dict[idx]
    elif isinstance(idx, int) and idx < len(G.DEVICE_LIST):
        current_dev = G.DEVICE_LIST[idx]
    else:
        raise IndexError("device idx not found in: %s or %s" % (
            list(dev_dict.keys()), list(range(len(G.DEVICE_LIST)))))
    G.DEVICE = current_dev


def auto_setup(basedir=None, devices=None, logdir=None, project_root=None, compress=None):
    """
    Auto setup running env and try connect android device if not device connected.

    :param basedir: basedir of script, __file__ is also acceptable.
    :param devices: connect_device uri in list.
    :param logdir: log dir for script report, default is None for no log, set to ``True`` for ``<basedir>/log``.
    :param project_root: project root dir for `using` api.
    :param compress: The compression rate of the screenshot image, integer in range [1, 99], default is 10
    :Example:
        >>> auto_setup(__file__)
        >>> auto_setup(__file__, devices=["Android://127.0.0.1:5037/SJE5T17B17"],
        ...             logdir=True, project_root=r"D:\\test\\logs", compress=90)
    """
    if basedir:
        if os.path.isfile(basedir):
            basedir = os.path.dirname(basedir)
        if basedir not in G.BASEDIR:
            G.BASEDIR.append(basedir)
    if logdir:
        logdir = script_log_dir(basedir, logdir)
        set_logdir(logdir)
    if devices:
        for dev in devices:
            connect_device(dev)
    if project_root:
        ST.PROJECT_ROOT = project_root
    if compress:
        ST.SNAPSHOT_QUALITY = compress


"""
Device Operations
"""


@logwrap
def shell(cmd):
    """
    Start remote shell in the target device and execute the command

    :param cmd: command to be run on device, e.g. "ls /data/local/tmp"
    :return: the output of the shell cmd
    :platforms: Android
    :Example:
        >>> # Execute commands on the current device adb shell ls
        >>> print(shell("ls"))

        >>> # Execute adb instructions for specific devices
        >>> dev = connect_device("Android:///device1")
        >>> dev.shell("ls")

        >>> # Switch to a device and execute the adb command
        >>> set_current(0)
        >>> shell("ls")
    """
    return G.DEVICE.shell(cmd)


@logwrap
def start_app(package, activity=None):
    """
    Start the target application on device

    :param package: name of the package to be started, e.g. "com.netease.my"
    :param activity: the activity to start, default is None which means the main activity
    :return: None
    :platforms: Android, iOS
    :Example:
        >>> start_app("com.netease.cloudmusic")
        >>> start_app("com.apple.mobilesafari")  # on iOS
    """
    G.DEVICE.start_app(package, activity)


@logwrap
def stop_app(package):
    """
    Stop the target application on device

    :param package: name of the package to stop, see also `start_app`
    :return: None
    :platforms: Android, iOS
    :Example:
        >>> stop_app("com.netease.cloudmusic")
    """
    G.DEVICE.stop_app(package)


@logwrap
def clear_app(package):
    """
    Clear data of the target application on device

    :param package: name of the package,  see also `start_app`
    :return: None
    :platforms: Android
    :Example:
        >>> clear_app("com.netease.cloudmusic")
    """
    G.DEVICE.clear_app(package)


@logwrap
def install(filepath, **kwargs):
    """
    Install application on device

    :param filepath: the path to file to be installed on target device
    :param kwargs: platform specific `kwargs`, please refer to corresponding docs
    :return: None
    :platforms: Android, iOS
    :Example:
        >>> install(r"D:\demo\test.apk")  # install Android apk
        >>> # adb install -r -t D:\\demo\\test.apk
        >>> install(r"D:\demo\test.apk", install_options=["-r", "-t"])

        >>> install(r"D:\demo\test.ipa") # install iOS ipa
        >>> install("http://www.example.com/test.ipa") # install iOS ipa from url

    """
    return G.DEVICE.install_app(filepath, **kwargs)


@logwrap
def uninstall(package):
    """
    Uninstall application on device

    :param package: name of the package, see also `start_app`
    :return: None
    :platforms: Android, iOS
    :Example:
        >>> uninstall("com.netease.cloudmusic")
    """
    return G.DEVICE.uninstall_app(package)


@logwrap
def snapshot(filename=None, msg="", quality=None, max_size=None):
    """
    Take the screenshot of the target device and save it to the file.

    :param filename: name of the file where to save the screenshot. If the relative path is provided, the default
                     location is ``ST.LOG_DIR``
    :param msg: short description for screenshot, it will be recorded in the report
    :param quality: The image quality, integer in range [1, 99], default is 10
    :param max_size: the maximum size of the picture, e.g 1200
    :return: {"screen": filename, "resolution": resolution of the screen} or None
    :platforms: Android, iOS, Windows
    :Example:
        >>> snapshot(msg="index")
        >>> # save the screenshot to test.jpg
        >>> snapshot(filename="test.png", msg="test")

        The quality and size of the screenshot can be set::

        >>> # Set the screenshot quality to 30
        >>> ST.SNAPSHOT_QUALITY = 30
        >>> # Set the screenshot size not to exceed 600*600
        >>> # if not set, the default size is the original image size
        >>> ST.IMAGE_MAXSIZE = 600
        >>> # The quality of the screenshot is 30, and the size does not exceed 600*600
        >>> touch((100, 100))
        >>> # The quality of the screenshot of this sentence is 90
        >>> snapshot(filename="test.png", msg="test", quality=90)
        >>> # The quality of the screenshot is 90, and the size does not exceed 1200*1200
        >>> snapshot(filename="test2.png", msg="test", quality=90, max_size=1200)

    """
    if not quality:
        quality = ST.SNAPSHOT_QUALITY
    if not max_size and ST.IMAGE_MAXSIZE:
        max_size = ST.IMAGE_MAXSIZE
    if filename:
        if not os.path.isabs(filename):
            logdir = ST.LOG_DIR or "."
            filename = os.path.join(logdir, filename)
        screen = G.DEVICE.snapshot(filename, quality=quality, max_size=max_size)
        return try_log_screen(screen, quality=quality, max_size=max_size)
    else:
        return try_log_screen(quality=quality, max_size=max_size)


@logwrap
def wake():
    """
    Wake up and unlock the target device

    :return: None
    :platforms: Android
    :Example:
        >>> wake()

    .. note:: Might not work on some models
    """
    G.DEVICE.wake()


@logwrap
def home():
    """
    Return to the home screen of the target device.

    :return: None
    :platforms: Android, iOS
    :Example:
        >>> home()
    """
    G.DEVICE.home()


@logwrap
def touch(v, times=1, **kwargs):
    """
    Perform the touch action on the device screen

    :param v: target to touch, either a ``Template`` instance or absolute coordinates (x, y)
    :param times: how many touches to be performed
    :param kwargs: platform specific `kwargs`, please refer to corresponding docs
    :return: finial position to be clicked, e.g. (100, 100)
    :platforms: Android, Windows, iOS
    :Example:
        Click absolute coordinates::

        >>> touch((100, 100))

        Click the center of the picture(Template object)::

        >>> touch(Template(r"tpl1606730579419.png", target_pos=5))

        Click on relative coordinates, for example, click on the center of the screen::

        >>> touch((0.5, 0.5))

        Click 2 times::

        >>> touch((100, 100), times=2)

        Under Android and Windows platforms, you can set the click duration::

        >>> touch((100, 100), duration=2)

        Right click(Windows)::

        >>> touch((100, 100), right_click=True)

    """
    if isinstance(v, Template):
        pos = loop_find(v, timeout=ST.FIND_TIMEOUT)
    else:
        try_log_screen()
        pos = v
    for _ in range(times):
        # If pos is a relative coordinate, return the converted click coordinates.
        # iOS may all use vertical screen coordinates, so coordinates will not be returned.
        pos = G.DEVICE.touch(pos, **kwargs) or pos
        time.sleep(0.05)
    delay_after_operation()
    return pos

click = touch  # click is alias of touch


@logwrap
def double_click(v):
    """
    Perform double click

    :param v: target to touch, either a ``Template`` instance or absolute coordinates (x, y)
    :return: finial position to be clicked
    :Example:

        >>> double_click((100, 100))
        >>> double_click(Template(r"tpl1606730579419.png"))
    """
    if isinstance(v, Template):
        pos = loop_find(v, timeout=ST.FIND_TIMEOUT)
    else:
        try_log_screen()
        pos = v
    pos = G.DEVICE.double_click(pos) or pos
    delay_after_operation()
    return pos


@logwrap
def swipe(v1, v2=None, vector=None, **kwargs):
    """
    Perform the swipe action on the device screen.

    There are two ways of assigning the parameters
        * ``swipe(v1, v2=Template(...))``   # swipe from v1 to v2
        * ``swipe(v1, vector=(x, y))``      # swipe starts at v1 and moves along the vector.


    :param v1: the start point of swipe,
               either a Template instance or absolute coordinates (x, y)
    :param v2: the end point of swipe,
               either a Template instance or absolute coordinates (x, y)
    :param vector: a vector coordinates of swipe action, either absolute coordinates (x, y) or percentage of
                   screen e.g.(0.5, 0.5)
    :param **kwargs: platform specific `kwargs`, please refer to corresponding docs
    :raise Exception: general exception when not enough parameters to perform swap action have been provided
    :return: Origin position and target position
    :platforms: Android, Windows, iOS
    :Example:

        >>> swipe(Template(r"tpl1606814865574.png"), vector=[-0.0316, -0.3311])
        >>> swipe((100, 100), (200, 200))

        Custom swiping duration and number of steps(Android and iOS)::

        >>> # swiping lasts for 1 second, divided into 6 steps
        >>> swipe((100, 100), (200, 200), duration=1, steps=6)

        Use relative coordinates to swipe, such as swiping from the center right to the left of the screen::

        >>> swipe((0.7, 0.5), (0.2, 0.5))

    """
    if isinstance(v1, Template):
        try:
            pos1 = loop_find(v1, timeout=ST.FIND_TIMEOUT)
        except TargetNotFoundError:
            # 如果由图1滑向图2，图1找不到，会导致图2的文件路径未被初始化，可能在报告中不能正确显示
            if v2 and isinstance(v2, Template):
                v2.filepath
            raise
    else:
        try_log_screen()
        pos1 = v1

    if v2:
        if isinstance(v2, Template):
            pos2 = loop_find(v2, timeout=ST.FIND_TIMEOUT_TMP)
        else:
            pos2 = v2
    elif vector:
        if vector[0] <= 1 and vector[1] <= 1:
            w, h = G.DEVICE.get_current_resolution()
            vector = (int(vector[0] * w), int(vector[1] * h))
        pos2 = (pos1[0] + vector[0], pos1[1] + vector[1])
    else:
        raise Exception("no enough params for swipe")

    pos1, pos2 = G.DEVICE.swipe(pos1, pos2, **kwargs) or (pos1, pos2)
    delay_after_operation()
    return pos1, pos2


@logwrap
def pinch(in_or_out='in', center=None, percent=0.5):
    """
    Perform the pinch action on the device screen

    :param in_or_out: pinch in or pinch out, enum in ["in", "out"]
    :param center: center of pinch action, default as None which is the center of the screen
    :param percent: percentage of the screen of pinch action, default is 0.5
    :return: None
    :platforms: Android
    :Example:

        Pinch in the center of the screen with two fingers::

        >>> pinch()

        Take (100,100) as the center and slide out with two fingers::

        >>> pinch('out', center=(100, 100))
    """
    try_log_screen()
    G.DEVICE.pinch(in_or_out=in_or_out, center=center, percent=percent)
    delay_after_operation()


@logwrap
def keyevent(keyname, **kwargs):
    """
    Perform key event on the device

    :param keyname: platform specific key name
    :param **kwargs: platform specific `kwargs`, please refer to corresponding docs
    :return: None
    :platforms: Android, Windows, iOS
    :Example:

        * ``Android``: it is equivalent to executing ``adb shell input keyevent KEYNAME`` ::

        >>> keyevent("HOME")
        >>> # The constant corresponding to the home key is 3
        >>> keyevent("3")  # same as keyevent("HOME")
        >>> keyevent("BACK")
        >>> keyevent("KEYCODE_DEL")

        .. seealso::

           Module :py:mod:`airtest.core.android.adb.ADB.keyevent`
              Equivalent to calling the ``android.adb.keyevent()``

           `Android Keyevent <https://developer.android.com/reference/android/view/KeyEvent#constants_1>`_
              Documentation for more ``Android.KeyEvent``

        * ``Windows``: Use ``pywinauto.keyboard`` module for key input::

        >>> keyevent("{DEL}")
        >>> keyevent("%{F4}")  # close an active window with Alt+F4

        .. seealso::

            Module :py:mod:`airtest.core.win.win.Windows.keyevent`

            `pywinauto.keyboard <https://pywinauto.readthedocs.io/en/latest/code/pywinauto.keyboard.html>`_
                Documentation for ``pywinauto.keyboard``

        * ``iOS``: Only supports home/volumeUp/volumeDown::

        >>> keyevent("HOME")
        >>> keyevent("volumeUp")

    """
    G.DEVICE.keyevent(keyname, **kwargs)
    delay_after_operation()


@logwrap
def text(text, enter=True, **kwargs):
    """
    Input text on the target device. Text input widget must be active first.

    :param text: text to input, unicode is supported
    :param enter: input `Enter` keyevent after text input, default is True
    :return: None
    :platforms: Android, Windows, iOS
    :Example:

        >>> text("test")
        >>> text("test", enter=False)

        On Android, sometimes you need to click the search button after typing::

        >>> text("test", search=True)

        .. seealso::

            Module :py:mod:`airtest.core.android.ime.YosemiteIme.code`

            If you want to enter other keys on the keyboard, you can use the interface::

                >>> text("test")
                >>> device().yosemite_ime.code("3")  # 3 = IME_ACTION_SEARCH

            Ref: `Editor Action Code <http://developer.android.com/reference/android/view/inputmethod/EditorInfo.html>`_

    """
    G.DEVICE.text(text, enter=enter, **kwargs)
    delay_after_operation()


@logwrap
def sleep(secs=1.0):
    """
    Set the sleep interval. It will be recorded in the report

    :param secs: seconds to sleep
    :return: None
    :platforms: Android, Windows, iOS
    :Example:

        >>> sleep(1)
    """
    time.sleep(secs)


@logwrap
def wait(v, timeout=None, interval=0.5, intervalfunc=None):
    """
    Wait to match the Template on the device screen

    :param v: target object to wait for, Template instance
    :param timeout: time interval to wait for the match, default is None which is ``ST.FIND_TIMEOUT``
    :param interval: time interval in seconds to attempt to find a match
    :param intervalfunc: called after each unsuccessful attempt to find the corresponding match
    :raise TargetNotFoundError: raised if target is not found after the time limit expired
    :return: coordinates of the matched target
    :platforms: Android, Windows, iOS
    :Example:

        >>> wait(Template(r"tpl1606821804906.png"))  # timeout after ST.FIND_TIMEOUT
        >>> # find Template every 3 seconds, timeout after 120 seconds
        >>> wait(Template(r"tpl1606821804906.png"), timeout=120, interval=3)

        You can specify a callback function every time the search target fails::

        >>> def notfound():
        >>>     print("No target found")
        >>> wait(Template(r"tpl1607510661400.png"), intervalfunc=notfound)

    """
    timeout = timeout or ST.FIND_TIMEOUT
    pos = loop_find(v, timeout=timeout, interval=interval, intervalfunc=intervalfunc)
    return pos


@logwrap
def exists(v):
    """
    Check whether given target exists on device screen

    :param v: target to be checked
    :return: False if target is not found, otherwise returns the coordinates of the target
    :platforms: Android, Windows, iOS
    :Example:

        >>> if exists(Template(r"tpl1606822430589.png")):
        >>>     touch(Template(r"tpl1606822430589.png"))

        Since ``exists()`` will return the coordinates, we can directly click on this return value to reduce one image search::

        >>> pos = exists(Template(r"tpl1606822430589.png"))
        >>> if pos:
        >>>     touch(pos)

    """
    try:
        pos = loop_find(v, timeout=ST.FIND_TIMEOUT_TMP)
    except TargetNotFoundError:
        return False
    else:
        return pos


@logwrap
def find_all(v):
    """
    Find all occurrences of the target on the device screen and return their coordinates

    :param v: target to find
    :return: list of results, [{'result': (x, y),
                                'rectangle': ( (left_top, left_bottom, right_bottom, right_top) ),
                                'confidence': 0.9},
                                ...]
    :platforms: Android, Windows, iOS
    :Example:

        >>> find_all(Template(r"tpl1607511235111.png"))
        [{'result': (218, 468), 'rectangle': ((149, 440), (149, 496), (288, 496), (288, 440)),
        'confidence': 0.9999996423721313}]

    """
    screen = G.DEVICE.snapshot(quality=ST.SNAPSHOT_QUALITY)
    return v.match_all_in(screen)


@logwrap
def get_clipboard(*args, **kwargs):
    """
    Get the content from the clipboard.

    :return: str
    :platforms: Android, iOS, Windows
    :Example:

        >>> text = get_clipboard()  # Android or local iOS
        >>> print(text)

        >>> # When the iOS device is a remote device, or more than one wda is installed on the device, you need to specify the wda_bundle_id
        >>> text = get_clipboard(wda_bundle_id="com.WebDriverAgentRunner.xctrunner")
        >>> print(text)

    """
    return G.DEVICE.get_clipboard(*args, **kwargs)


@logwrap
def set_clipboard(content, *args, **kwargs):
    """
    Set the content from the clipboard.

    :param content: str
    :return: None
    :platforms: Android, iOS, Windows
    :Example:

        >>> set_clipboard("content")  # Android or local iOS
        >>> print(get_clipboard())

        >>> # When the iOS device is a remote device, or more than one wda is installed on the device, you need to specify the wda_bundle_id
        >>> set_clipboard("content", wda_bundle_id="com.WebDriverAgentRunner.xctrunner")

    """
    G.DEVICE.set_clipboard(content, *args, **kwargs)


@logwrap
def paste(*args, **kwargs):
    """
    Paste the content from the clipboard.

    :platforms: Android, iOS, Windows
    :return: None
    :Example:

        >>> set_clipboard("content")
        >>> paste()  # will paste "content" to the device

    """
    G.DEVICE.paste(*args, **kwargs)


"""
Assertions: see airtest/core/assertions.py
"""



# -*- coding: utf-8 -*-

import unittest
import os
import sys
import six
import re
import shutil
import traceback
import warnings
from io import open
from airtestProject.airtest.core.api import G, auto_setup, log
from airtestProject.airtest.core.settings import Settings as ST
from airtestProject.airtest.utils.compat import decode_path, script_dir_name, script_log_dir
from copy import copy


class AirtestCase(unittest.TestCase):

    PROJECT_ROOT = "."
    SCRIPTEXT = ".air"
    TPLEXT = ".png"

    @classmethod
    def setUpClass(cls):
        cls.args = args

        setup_by_args(args)

        # setup script exec scope
        cls.scope = copy(globals())
        cls.scope["exec_script"] = cls.exec_other_script

    def setUp(self):
        if self.args.log and self.args.recording:
            for dev in G.DEVICE_LIST:
                try:
                    dev.start_recording()
                except:
                    traceback.print_exc()
        # support for if __name__ == '__main__'
        self.scope['__name__'] = '__main__'

    def tearDown(self):
        if self.args.log and self.args.recording:
            for k, dev in enumerate(G.DEVICE_LIST):
                # 录屏文件保存的命名规则：
                # 如果未指定文件名，只传了--recording，就默认用recording_手机序列号.mp4来命名
                # 如果指定了文件名--recording test.mp4，且超过一台手机，就命名为 手机序列号_test.mp4
                # 否则直接用指定的文件名保存录屏，必须是mp4结尾
                try:
                    if isinstance(self.args.recording, six.string_types) and self.args.recording.endswith(".mp4"):
                        basename = os.path.basename(self.args.recording)
                        output_name = dev.serialno + "_" + basename if len(G.DEVICE_LIST) > 1 else basename
                    else:
                        output_name = "recording_{serialno}.mp4".format(serialno=dev.serialno)
                    output = os.path.join(self.args.log, output_name)
                    dev.stop_recording(output)
                except:
                    traceback.print_exc()

    def runTest(self):
        scriptpath, pyfilename = script_dir_name(self.args.script)
        pyfilepath = os.path.join(scriptpath, pyfilename)
        pyfilepath = os.path.abspath(pyfilepath)
        self.scope["__file__"] = pyfilepath
        with open(pyfilepath, 'r', encoding="utf8") as f:
            code = f.read()
        pyfilepath = pyfilepath.encode(sys.getfilesystemencoding())

        try:
            exec(compile(code.encode("utf-8"), pyfilepath, 'exec'), self.scope)
        except Exception as err:
            log(err, desc="Final Error", snapshot=True if G.DEVICE_LIST else False)
            six.reraise(*sys.exc_info())

    @classmethod
    def exec_other_script(cls, scriptpath):
        """run other script in test script"""

        warnings.simplefilter("always")
        warnings.warn("please use using() api instead.", PendingDeprecationWarning)

        def _sub_dir_name(scriptname):
            dirname = os.path.splitdrive(os.path.normpath(scriptname))[-1]
            dirname = dirname.strip(os.path.sep).replace(os.path.sep, "_").replace(cls.SCRIPTEXT, "_sub")
            return dirname

        def _copy_script(src, dst):
            if os.path.isdir(dst):
                shutil.rmtree(dst, ignore_errors=True)
            os.mkdir(dst)
            for f in os.listdir(src):
                srcfile = os.path.join(src, f)
                if not (os.path.isfile(srcfile) and f.endswith(cls.TPLEXT)):
                    continue
                dstfile = os.path.join(dst, f)
                shutil.copy(srcfile, dstfile)

        # find script in PROJECT_ROOT
        scriptpath = os.path.join(ST.PROJECT_ROOT, scriptpath)
        # copy submodule's images into sub_dir
        sub_dir = _sub_dir_name(scriptpath)
        sub_dirpath = os.path.join(cls.args.script, sub_dir)
        _copy_script(scriptpath, sub_dirpath)
        # read code
        pyfilename = os.path.basename(scriptpath).replace(cls.SCRIPTEXT, ".py")
        pyfilepath = os.path.join(scriptpath, pyfilename)
        pyfilepath = os.path.abspath(pyfilepath)
        with open(pyfilepath, 'r', encoding='utf8') as f:
            code = f.read()
        # replace tpl filepath with filepath in sub_dir
        code = re.sub("[\'\"](\w+.png)[\'\"]", "\"%s/\g<1>\"" % sub_dir, code)
        exec(compile(code.encode("utf8"), pyfilepath, 'exec'), cls.scope)


def setup_by_args(args):
    # init devices
    if isinstance(args.device, list):
        devices = args.device
    elif args.device:
        devices = [args.device]
    else:
        devices = []
        print("do not connect device")

    # set base dir to find tpl
    dirpath, _ = script_dir_name(args.script)

    # set log dir
    if args.log:
        args.log = script_log_dir(dirpath, args.log)
        print("save log in '%s'" % args.log)
    else:
        print("do not save log")

    # set snapshot quality
    if args.compress:
        compress = args.compress
    else:
        compress = ST.SNAPSHOT_QUALITY

    if args.no_image:
        ST.SAVE_IMAGE = False

    # guess project_root to be basedir of current .air path
    project_root = os.path.dirname(args.script) if not ST.PROJECT_ROOT else None

    auto_setup(dirpath, devices, args.log, project_root, compress)


def run_script(parsed_args, testcase_cls=AirtestCase):
    global args  # make it global deliberately to be used in AirtestCase & test scripts
    args = parsed_args
    suite = unittest.TestSuite()
    suite.addTest(testcase_cls())
    result = unittest.TextTestRunner(verbosity=0).run(suite)
    if not result.wasSuccessful():
        if result.failures and "AssertionError" in repr(result.failures):
            sys.exit(20)
        sys.exit(-1)



# -*- coding: utf-8 -*-
import os
import re
import sys
import time
import random
import platform
import psutil
import warnings
import subprocess
import threading
from copy import copy
from six import PY3, text_type, binary_type
from six.moves import reduce

from airtestProject.airtest.core.android.constant import (DEFAULT_ADB_PATH, IP_PATTERN,
                                           SDK_VERISON_ANDROID7)
from airtestProject.airtest.core.error import (AdbError, AdbShellError, AirtestError,
                                DeviceConnectionError)
from airtestProject.airtest.utils.compat import decode_path, raisefrom, proc_communicate_timeout, SUBPROCESS_FLAG
from airtestProject.airtest.utils.logger import get_logger
from airtestProject.airtest.utils.nbsp import NonBlockingStreamReader
from airtestProject.airtest.utils.retry import retries
from airtestProject.airtest.utils.apkparser import APK
from airtestProject.airtest.utils.snippet import get_std_encoding, reg_cleanup, split_cmd, make_file_executable

LOGGING = get_logger(__name__)


class ADB(object):
    """adb client object class"""

    _instances = []
    status_device = "device"
    status_offline = "offline"
    SHELL_ENCODING = "utf-8"

    def __init__(self, serialno=None, adb_path=None, server_addr=None, display_id=None, input_event=None):
        self.serialno = serialno
        self.adb_path = adb_path or self.get_adb_path()
        self.display_id = display_id
        self.input_event = input_event
        self._set_cmd_options(server_addr)
        self.connect()
        self._sdk_version = None
        self._line_breaker = None
        self._display_info = {}
        self._display_info_lock = threading.Lock()
        self._forward_local_using = []
        self.__class__._instances.append(self)

    @staticmethod
    def get_adb_path():
        """
        Returns the path to the adb executable.

        Checks if adb process is running, returns the running process path.

        If adb process is not running, checks if ANDROID_HOME environment variable is set.
        Constructs the adb path using ANDROID_HOME and returns it if set.

        If adb process is not running and ANDROID_HOME is not set, uses built-in adb path.

        Returns:
            str: The path to the adb executable.
        """
        if platform.system() == "Windows":
            ADB_NAME = "adb.exe"
        else:
            ADB_NAME = "adb"

        # Check if adb process is already running
        try:
            for process in psutil.process_iter(['name', 'exe']):
                if process.info['name'] == ADB_NAME and process.info['exe'] and os.path.exists(process.info['exe']):
                    return process.info['exe']
        except:
            # maybe OSError
            pass

        # Check if ANDROID_HOME environment variable exists
        android_home = os.environ.get('ANDROID_HOME')
        if android_home:
            adb_path = os.path.join(android_home, 'platform-tools', ADB_NAME)
            if os.path.exists(adb_path):
                return adb_path

        # Use airtest builtin adb path
        builtin_adb_path = ADB.builtin_adb_path()
        return builtin_adb_path

    @staticmethod
    def builtin_adb_path():
        """
        Return built-in adb executable path

        Returns:
            adb executable path

        """
        system = platform.system()
        machine = platform.machine()
        adb_path = DEFAULT_ADB_PATH.get('{}-{}'.format(system, machine))
        if not adb_path:
            adb_path = DEFAULT_ADB_PATH.get(system)
        if not adb_path:
            raise RuntimeError("No adb executable supports this platform({}-{}).".format(system, machine))

        if system != "Windows":
            # chmod +x adb
            make_file_executable(adb_path)
        return adb_path

    def _set_cmd_options(self, server_addr=None):
        """
        Set communication parameters (host and port) between adb server and adb client

        Args:
            server_addr: adb server address, default is ["127.0.0.1", 5037]

        Returns:
            None

        """
        self.host = server_addr[0] if server_addr else "127.0.0.1"
        self.port = server_addr[1] if server_addr else 5037
        self.cmd_options = [self.adb_path]
        if self.host not in ("localhost", "127.0.0.1"):
            self.cmd_options += ['-H', self.host]
        if self.port != 5037:
            self.cmd_options += ['-P', str(self.port)]

    def start_server(self):
        """
        Perform `adb start-server` command to start the adb server

        Returns:
            None

        """
        return self.cmd("start-server", device=False)

    def kill_server(self):
        """
        Perform `adb kill-server` command to kill the adb server

        Returns:
            None

        """
        return self.cmd("kill-server", device=False)

    def version(self):
        """
        Perform `adb version` command and return the command output

        Returns:
            command output

        """
        return self.cmd("version", device=False).strip()

    def start_cmd(self, cmds, device=True):
        """
        Start a subprocess with adb command(s)

        Args:
            cmds: command(s) to be run
            device: if True, the device serial number must be specified by `-s serialno` argument

        Raises:
            RuntimeError: if `device` is True and serialno is not specified

        Returns:
            a subprocess

        """
        if device:
            if not self.serialno:
                raise RuntimeError("please set serialno first")
            cmd_options = self.cmd_options + ['-s', self.serialno]
        else:
            cmd_options = self.cmd_options

        cmds = cmd_options + split_cmd(cmds)
        LOGGING.debug(" ".join(cmds))

        if not PY3:
            cmds = [c.encode(get_std_encoding(sys.stdin)) for c in cmds]

        proc = subprocess.Popen(
            cmds,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            creationflags=SUBPROCESS_FLAG
        )
        return proc

    def cmd(self, cmds, device=True, ensure_unicode=True, timeout=None):
        """
        Run the adb command(s) in subprocess and return the standard output

        Args:
            cmds: command(s) to be run
            device: if True, the device serial number must be specified by -s serialno argument
            ensure_unicode: encode/decode unicode of standard outputs (stdout, stderr)
            timeout: timeout in seconds

        Raises:
            DeviceConnectionError: if any error occurs when connecting the device
            AdbError: if any other adb error occurs

        Returns:
            command(s) standard output (stdout)

        """
        proc = self.start_cmd(cmds, device)
        if timeout:
            stdout, stderr = proc_communicate_timeout(proc, timeout)
        else:
            stdout, stderr = proc.communicate()

        if ensure_unicode:
            stdout = stdout.decode(get_std_encoding(sys.stdout))
            stderr = stderr.decode(get_std_encoding(sys.stderr))

        if proc.returncode > 0:
            # adb connection error
            pattern = DeviceConnectionError.DEVICE_CONNECTION_ERROR
            if isinstance(stderr, binary_type):
                pattern = pattern.encode("utf-8")
            if re.search(pattern, stderr):
                raise DeviceConnectionError(stderr)
            else:
                raise AdbError(stdout, stderr)
        return stdout

    def close_proc_pipe(self, proc):
        """close stdin/stdout/stderr of subprocess.Popen."""

        def close_pipe(pipe):
            if pipe:
                pipe.close()

        close_pipe(proc.stdin)
        close_pipe(proc.stdout)
        close_pipe(proc.stderr)

    def devices(self, state=None):
        """
        Perform `adb devices` command and return the list of adb devices

        Args:
            state: optional parameter to filter devices in specific state

        Returns:
            list od adb devices

        """
        patten = re.compile(r'[\w\d.:-]+\t[\w]+$')
        device_list = []
        # self.start_server()
        output = self.cmd("devices", device=False)
        for line in output.splitlines():
            line = line.strip()
            if not line or not patten.match(line):
                continue
            serialno, cstate = line.split('\t')
            if state and cstate != state:
                continue
            device_list.append((serialno, cstate))
        return device_list

    def connect(self, force=False):
        """
        Perform `adb connect` command, remote devices are preferred to connect first

        Args:
            force: force connection, default is False

        Returns:
            None

        """
        if self.serialno and ":" in self.serialno and (force or self.get_status() != "device"):
            connect_result = self.cmd("connect %s" % self.serialno)
            LOGGING.info(connect_result)

    def disconnect(self):
        """
        Perform `adb disconnect` command

        Returns:
            None

        """
        if ":" in self.serialno:
            self.cmd("disconnect %s" % self.serialno)

    def get_status(self):
        """
        Perform `adb get-state` and return the device status

        Raises:
            AdbError: if status cannot be obtained from the device

        Returns:
            None if status is `not found`, otherwise return the standard output from `adb get-state` command

        """
        proc = self.start_cmd("get-state")
        stdout, stderr = proc.communicate()

        stdout = stdout.decode(get_std_encoding(sys.stdout))
        stderr = stderr.decode(get_std_encoding(sys.stdout))

        if proc.returncode == 0:
            return stdout.strip()
        elif "not found" in stderr:
            return None
        else:
            raise AdbError(stdout, stderr)

    def wait_for_device(self, timeout=5):
        """
        Perform `adb wait-for-device` command

        Args:
            timeout: time interval in seconds to wait for device

        Raises:
            DeviceConnectionError: if device is not available after timeout

        Returns:
            None

        """
        try:
            self.cmd("wait-for-device", timeout=timeout)
        except RuntimeError as e:
            raisefrom(DeviceConnectionError, "device not ready", e)

    def start_shell(self, cmds):
        """
        Handle `adb shell` command(s)

        Args:
            cmds: adb shell command(s)

        Returns:
            None

        """
        cmds = ['shell'] + split_cmd(cmds)
        return self.start_cmd(cmds)

    def raw_shell(self, cmds, ensure_unicode=True):
        """
        Handle `adb shell` command(s) with unicode support

        Args:
            cmds: adb shell command(s)
            ensure_unicode: decode/encode unicode True or False, default is True

        Returns:
            command(s) output

        """
        cmds = ['shell'] + split_cmd(cmds)
        out = self.cmd(cmds, ensure_unicode=False)
        if not ensure_unicode:
            return out
        # use shell encoding to decode output
        try:
            return out.decode(self.SHELL_ENCODING)
        except UnicodeDecodeError:
            warnings.warn("shell output decode {} fail. repr={}".format(self.SHELL_ENCODING, repr(out)))
            return text_type(repr(out))

    def shell(self, cmd):
        """
        Run the `adb shell` command on the device

        Args:
            cmd: a command to be run

        Raises:
            AdbShellError: if command return value is non-zero or if any other `AdbError` occurred

        Returns:
            command output

        """
        if self.sdk_version < SDK_VERISON_ANDROID7:
            # for sdk_version < 25, adb shell do not raise error
            # https://stackoverflow.com/questions/9379400/adb-error-codes
            cmd = split_cmd(cmd) + [";", "echo", "---$?---"]
            out = self.raw_shell(cmd).rstrip()
            m = re.match("(.*)---(\d+)---$", out, re.DOTALL)
            if not m:
                warnings.warn("return code not matched")
                stdout = out
                returncode = 0
            else:
                stdout = m.group(1)
                returncode = int(m.group(2))
            if returncode > 0:
                raise AdbShellError("", stdout)
            return stdout
        else:
            try:
                out = self.raw_shell(cmd)
            except AdbError as err:
                raise AdbShellError(err.stdout, err.stderr)
            else:
                return out

    def keyevent(self, keyname):
        """
        Perform `adb shell input keyevent` command on the device

        Args:
            keyname: key event name

        Returns:
            None

        """
        self.shell(["input", "keyevent", keyname.upper()])

    def getprop(self, key, strip=True):
        """
        Perform `adb shell getprop` on the device

        Args:
            key: key value for property
            strip: True or False to strip the return carriage and line break from returned string

        Returns:
            propery value

        """
        prop = self.raw_shell(['getprop', key])
        if strip:
            if "\r\r\n" in prop:
                # Some mobile phones will output multiple lines of extra log
                prop = prop.split("\r\r\n")
                if len(prop) > 1:
                    prop = prop[-2]
                else:
                    prop = prop[-1]
            else:
                prop = prop.strip("\r\n")
        return prop

    @property
    @retries(max_tries=3)
    def sdk_version(self):
        """
        Get the SDK version from the device

        Returns:
            SDK version
        """
        if self._sdk_version is None:
            keyname = 'ro.build.version.sdk'
            self._sdk_version = int(self.getprop(keyname))
        return self._sdk_version

    def push(self, local, remote):
        """
        Perform `adb push` command

        Note:
            If there is a space (or special symbol) in the file name, it will be forced to add escape characters,
            and the new file name will be added with quotation marks and returned as the return value

            注意：文件名中如果带有空格（或特殊符号），将会被强制增加转义符，并将新的文件名添加引号，作为返回值返回

        Args:
            local: local file to be copied to the device
            remote: destination on the device where the file will be copied

        Returns:
            The file path saved in the phone may be enclosed in quotation marks, eg. '"test\ file.txt"'

        Examples:

            >>> adb = device().adb
            >>> adb.push("test.txt", "/data/local/tmp")

            >>> new_name = adb.push("test space.txt", "/data/local/tmp")  # test the space in file name
            >>> print(new_name)
            "/data/local/tmp/test\ space.txt"
            >>> adb.shell("rm " + new_name)

        """
        local = decode_path(local)  # py2
        if os.path.isfile(local) and os.path.splitext(local)[-1] != os.path.splitext(remote)[-1]:
            # If remote is a folder, add the filename and escape
            filename = os.path.basename(local)
            # Add escape characters for spaces, parentheses, etc. in filenames
            filename = re.sub(r"[ \(\)\&]", lambda m: "\\" + m.group(0), filename)
            remote = '%s/%s' % (remote, filename)
        self.cmd(["push", local, remote], ensure_unicode=False)
        return '\"%s\"' % remote

    def pull(self, remote, local):
        """
        Perform `adb pull` command

        Args:
            remote: remote file to be downloaded from the device
            local: local destination where the file will be downloaded from the device

        Note:
            If <=PY3, the path in Windows cannot be the root directory, and cannot contain symbols such as /g in the path
            注意：如果低于PY3,windows中路径不能为根目录，并且不能包含/g等符号在路径里

        Returns:
            None
        """
        local = decode_path(local)  # py2
        if PY3:
            # If python3, use Path to force / convert to \
            from pathlib import Path
            local = Path(local).as_posix()
        self.cmd(["pull", remote, local], ensure_unicode=False)

    def forward(self, local, remote, no_rebind=True):
        """
        Perform `adb forward` command

        Args:
            local: local tcp port to be forwarded
            remote: tcp port of the device where the local tcp port will be forwarded
            no_rebind: True or False

        Returns:
            None

        """
        cmds = ['forward']
        if no_rebind:
            cmds += ['--no-rebind']
        self.cmd(cmds + [local, remote])
        # register for cleanup atexit
        if local not in self._forward_local_using:
            self._forward_local_using.append(local)

    def get_forwards(self):
        """
        Perform `adb forward --list`command

        Yields:
            serial number, local tcp port, remote tcp port

        Returns:
            None

        """
        out = self.cmd(['forward', '--list'])
        for line in out.splitlines():
            line = line.strip()
            if not line:
                continue
            cols = line.split()
            if len(cols) != 3:
                continue
            serialno, local, remote = cols
            yield serialno, local, remote

    @classmethod
    def get_available_forward_local(cls):
        """
        Generate a pseudo random number between 11111 and 20000 that will be used as local forward port

        Returns:
            integer between 11111 and 20000

        Note:
            use `forward --no-rebind` to check if port is available
        """
        return random.randint(11111, 20000)

    @retries(3)
    def setup_forward(self, device_port, no_rebind=True):
        """
        Generate pseudo random local port and check if the port is available.

        Args:
            device_port: it can be string or the value of the `function(localport)`,
                         e.g. `"tcp:5001"` or `"localabstract:{}".format`
            no_rebind: adb forward --no-rebind option

        Returns:
            local port and device port

        """
        localport = self.get_available_forward_local()
        if callable(device_port):
            device_port = device_port(localport)
        self.forward("tcp:%s" % localport, device_port, no_rebind=no_rebind)
        return localport, device_port

    def remove_forward(self, local=None):
        """
        Perform `adb forward --remove` command

        Args:
            local: local tcp port

        Returns:
            None

        """
        if local:
            cmds = ["forward", "--remove", local]
        else:
            cmds = ["forward", "--remove-all"]
        try:
            self.cmd(cmds)
        except AdbError as e:
            # ignore if already removed or disconnected
            if "not found" in (repr(e.stdout) + repr(e.stderr)):
                pass
        except DeviceConnectionError:
            pass
        # unregister for cleanup
        if local in self._forward_local_using:
            self._forward_local_using.remove(local)

    def install_app(self, filepath, replace=False, install_options=None):
        """
        Perform `adb install` command

        Args:
            filepath: full path to file to be installed on the device
            replace: force to replace existing application, default is False

                e.g.["-t",  # allow test packages
                    "-l",  # forward lock application,
                    "-s",  # install application on sdcard,
                    "-d",  # allow version code downgrade (debuggable packages only)
                    "-g",  # grant all runtime permissions
                ]

        Returns:
            command output

        """
        if isinstance(filepath, str):
            filepath = decode_path(filepath)

        if not os.path.isfile(filepath):
            raise RuntimeError("file: %s does not exists" % (repr(filepath)))

        if not install_options or type(install_options) != list:
            install_options = []
        if replace:
            install_options.append("-r")
        cmds = ["install", ] + install_options + [filepath, ]
        try:
            out = self.cmd(cmds)
        except AdbError as e:
            out = repr(e.stderr) + repr(e.stdout)
            # If the signatures are inconsistent, uninstall the old version first
            if "INSTALL_FAILED_UPDATE_INCOMPATIBLE" in out and replace:
                try:
                    package_name = re.search(r"package (.*?) .*", out).group(1)
                except:
                    # get package name
                    package_name = APK(filepath).get_package()
                self.uninstall_app(package_name)
                out = self.cmd(cmds)
            else:
                raise

        if re.search(r"Failure \[.*?\]", out):
            raise AdbShellError("Installation Failure", repr(out))

        return out

    def install_multiple_app(self, filepath, replace=False, install_options=None):
        """
            Perform `adb install-multiple` command

            Args:
                filepath: full path to file to be installed on the device
                replace: force to replace existing application, default is False
                install_options:  list of options
                    e.g.["-t",  # allow test packages
                        "-l",  # forward lock application,
                        "-s",  # install application on sdcard,
                        "-d",  # allow version code downgrade (debuggable packages only)
                        "-g",  # grant all runtime permissions
                        "-p",  # partial application install (install-multiple only)
                    ]

            Returns:
                command output
        """
        if isinstance(filepath, str):
            filepath = decode_path(filepath)

        if not os.path.isfile(filepath):
            raise RuntimeError("file: %s does not exists" % (repr(filepath)))

        if not install_options or type(install_options) != list:
            install_options = []
        if replace:
            install_options.append("-r")
        cmds = ["install-multiple", ] + install_options + [filepath, ]

        try:
            out = self.cmd(cmds)
        except AdbError as err:
            if "Failed to finalize session".lower() in err.stderr.lower():
                return "Success"
            else:
                return self.install_app(filepath, replace)

        if re.search(r"Failure \[.*?\]", out):
            raise AdbShellError("Installation Failure", repr(out))

        return out

    def pm_install(self, filepath, replace=False, install_options=None):
        """
        Perform `adb push` and `adb install` commands

        Note:
            This is more reliable and recommended way of installing `.apk` files

        Args:
            filepath: full path to file to be installed on the device
            replace: force to replace existing application, default is False
            install_options:  list of options
                e.g.["-t",  # allow test packages
                    "-l",  # forward lock application,
                    "-s",  # install application on sdcard,
                    "-d",  # allow version code downgrade (debuggable packages only)
                    "-g",  # grant all runtime permissions
                ]

        Returns:
            None

        """
        if isinstance(filepath, str):
            filepath = decode_path(filepath)
        if not os.path.isfile(filepath):
            raise RuntimeError("file: %s does not exists" % (repr(filepath)))

        if not install_options or type(install_options) != list:
            install_options = []
        if replace:
            install_options.append("-r")

        filename = os.path.basename(filepath)
        device_dir = "/data/local/tmp"

        try:
            # if the apk file path contains spaces, the path must be escaped
            device_path = self.push(filepath, device_dir)
        except AdbError as err:
            # Error: no space left on device
            raise err

        try:
            cmds = ["pm", "install", ] + install_options + [device_path]
            self.shell(cmds)
        except AdbError as e:
            out = repr(e.stderr) + repr(e.stdout)
            # If the signatures are inconsistent, uninstall the old version first
            if re.search(r"INSTALL_FAILED_UPDATE_INCOMPATIBLE", out):
                try:
                    package_name = re.search(r"package (.*?) .*", out).group(1)
                except:
                    # get package name
                    package_name = APK(filepath).get_package()
                self.uninstall_app(package_name)
                out = self.shell(cmds)
            else:
                raise
        finally:
            # delete apk file
            self.cmd(["shell", "rm", device_path], timeout=30)

    def uninstall_app(self, package):
        """
        Perform `adb uninstall` command
        Args:
            package: package name to be uninstalled from the device

        Returns:
            command output

        """
        return self.cmd(['uninstall', package])

    def pm_uninstall(self, package, keepdata=False):
        """
        Perform `adb uninstall` command and delete all related application data

        Args:
            package: package name to be uninstalled from the device
            keepdata: True or False, keep application data after removing the app from the device

        Returns:
            command output

        """
        cmd = ['pm', 'uninstall', package]
        if keepdata:
            cmd.append('-k')
        self.shell(cmd)

    def snapshot(self):
        """
        Take the screenshot of the device display

        Returns:
            command output (stdout)

        """
        if self.display_id:
            raw = self.cmd('shell screencap -d {0} -p'.format(self.display_id), ensure_unicode=False)
        else:
            raw = self.cmd('shell screencap -p', ensure_unicode=False)
        return raw.replace(self.line_breaker, b"\n")

    # PEP 3113 -- Removal of Tuple Parameter Unpacking
    # https://www.python.org/dev/peps/pep-3113/
    def touch(self, tuple_xy):
        """
        Perform user input (touchscreen) on given coordinates

        Args:
            tuple_xy: coordinates (x, y)

        Returns:
            None

        """
        x, y = tuple_xy
        self.shell('input tap %d %d' % (x, y))
        time.sleep(0.1)

    def swipe(self, tuple_x0y0, tuple_x1y1, duration=500):
        """
        Perform user input (swipe screen) from start point (x,y) to end point (x,y)

        Args:
            tuple_x0y0: start point coordinates (x, y)
            tuple_x1y1: end point coordinates (x, y)
            duration: time interval for action, default 500

        Raises:
            AirtestError: if SDK version is not supported

        Returns:
            None

        """
        # prot python 3
        x0, y0 = tuple_x0y0
        x1, y1 = tuple_x1y1

        version = self.sdk_version
        if version <= 15:
            raise AirtestError('swipe: API <= 15 not supported (version=%d)' % version)
        elif version <= 17:
            self.shell('input swipe %d %d %d %d' % (x0, y0, x1, y1))
        else:
            self.shell('input touchscreen swipe %d %d %d %d %d' % (x0, y0, x1, y1, duration))

    def logcat(self, grep_str="", extra_args="", read_timeout=10):
        """
        Perform `adb shell logcat` command and search for given patterns

        Args:
            grep_str: pattern to filter from the logcat output
            extra_args: additional logcat arguments
            read_timeout: time interval to read the logcat, default is 10

        Yields:
            logcat lines containing filtered patterns

        Returns:
            None

        """
        cmds = "shell logcat"
        if extra_args:
            cmds += " " + extra_args
        if grep_str:
            cmds += " | grep " + grep_str
        logcat_proc = self.start_cmd(cmds)
        nbsp = NonBlockingStreamReader(logcat_proc.stdout, print_output=False)
        while True:
            line = nbsp.readline(read_timeout)
            if line is None:
                break
            else:
                yield line
        nbsp.kill()
        logcat_proc.kill()
        return

    def exists_file(self, filepath):
        """
        Check if the file exits on the device

        Args:
            filepath: path to the file

        Returns:
            True or False if file found or not

        """
        try:
            out = self.shell(["ls", filepath])
        except AdbShellError:
            return False
        else:
            return not ("No such file or directory" in out)

    def file_size(self, filepath):
        """
        Get the file size

        Args:
            filepath: path to the file

        Returns:
            The file size

        Raises:
            AdbShellError if no such file
        """
        out = self.shell(["ls", "-l", filepath])
        try:
            file_size = int(out.split()[4])
        except ValueError:
            # 安卓6.0.1系统得到的结果是[3]为文件大小
            file_size = int(out.split()[3])
        return file_size

    def _cleanup_forwards(self):
        """
        Remove the local forward ports

        Returns:
            None
        """
        # remove forward成功后，会改变self._forward_local_using的内容，因此需要copy一遍
        # After remove_forward() is successful, self._forward_local_using will be changed, so it needs to be copied
        forward_local_list = copy(self._forward_local_using)
        for local in forward_local_list:
            try:
                self.remove_forward(local)
            except DeviceConnectionError:
                # if device is offline, ignore
                pass

    @property
    def line_breaker(self):
        """
        Set carriage return and line break property for various platforms and SDK versions

        Returns:
            carriage return and line break string

        """
        if not self._line_breaker:
            if self.sdk_version >= SDK_VERISON_ANDROID7:
                line_breaker = os.linesep
            else:
                line_breaker = '\r' + os.linesep
            self._line_breaker = line_breaker.encode("ascii")
        return self._line_breaker

    @property
    def display_info(self):
        """
        Set device display properties (orientation, rotation and max values for x and y coordinates)

        Notes:
        if there is a lock screen detected, the function tries to unlock the device first

        Returns:
            device screen properties

        """
        self._display_info_lock.acquire()
        if not self._display_info:
            self._display_info = self.get_display_info()
        self._display_info_lock.release()
        return self._display_info

    def get_display_info(self):
        """
        Get information about device physical display (orientation, rotation and max values for x and y coordinates)

        Returns:
            device screen properties
            e.g {
                'width': 1440,
                'height': 2960,
                'density': 4.0,
                'orientation': 3,
                'rotation': 270,
                'max_x': 4095,
                'max_y': 4095
            }

        """
        display_info = self.getPhysicalDisplayInfo()
        orientation = self.getDisplayOrientation()
        max_x, max_y = self.getMaxXY()
        display_info.update({
            "orientation": orientation,
            "rotation": orientation * 90,
            "max_x": max_x,
            "max_y": max_y,
        })
        return display_info

    def getMaxXY(self):
        """
        Get device display maximum values for x and y coordinates

        Returns:
            max x and max y coordinates

        """
        ret = self.shell('getevent -p').split('\n')
        max_x, max_y = None, None
        for i in ret:
            if i.find("0035") != -1:
                patten = re.compile(r'max [0-9]+')
                ret = patten.search(i)
                if ret:
                    max_x = int(ret.group(0).split()[1])

            if i.find("0036") != -1:
                patten = re.compile(r'max [0-9]+')
                ret = patten.search(i)
                if ret:
                    max_y = int(ret.group(0).split()[1])
        return max_x, max_y

    def getRestrictedScreen(self):
        """
        Get value for mRestrictedScreen (without black border / virtual keyboard)`

        Returns:
            screen resolution mRestrictedScreen value as tuple (x, y)

        """
        # get the effective screen resolution of the device
        result = None
        # get the corresponding mRestrictedScreen parameters according to the device serial number
        dumpsys_info = self.shell("dumpsys window")
        match = re.search(r'mRestrictedScreen=.+', dumpsys_info)
        if match:
            infoline = match.group(0).strip()  # like 'mRestrictedScreen=(0,0) 720x1184'
            resolution = infoline.split(" ")[1].split("x")
            if isinstance(resolution, list) and len(resolution) == 2:
                result = int(str(resolution[0])), int(str(resolution[1]))

        return result

    def getPhysicalDisplayInfo(self):
        """
        Get value for display dimension and density from `mPhysicalDisplayInfo` value obtained from `dumpsys` command.

        Returns:
            physical display info for dimension and density

        """
        # use adb shell wm size
        displayInfo = {}
        try:
            wm_size = re.search(r'(?P<width>\d+)x(?P<height>\d+)\s*$', self.cmd('shell wm size', timeout=5))
        except (AdbError, RuntimeError) as e:
            LOGGING.error(e)
        else:
            if wm_size:
                displayInfo = dict((k, int(v)) for k, v in wm_size.groupdict().items())
                displayInfo['density'] = self._getDisplayDensity(strip=True)
                return displayInfo

        phyDispRE = re.compile('.*PhysicalDisplayInfo{(?P<width>\d+) x (?P<height>\d+), .*, density (?P<density>[\d.]+).*')
        out = self.raw_shell('dumpsys display')
        m = phyDispRE.search(out)
        if m:
            for prop in ['width', 'height']:
                displayInfo[prop] = int(m.group(prop))
            for prop in ['density']:
                # In mPhysicalDisplayInfo density is already a factor, no need to calculate
                displayInfo[prop] = float(m.group(prop))
            return displayInfo

        # This could also be mSystem or mOverscanScreen
        phyDispRE = re.compile('\s*mUnrestrictedScreen=\((?P<x>\d+),(?P<y>\d+)\) (?P<width>\d+)x(?P<height>\d+)')
        # This is known to work on older versions (i.e. API 10) where mrestrictedScreen is not available
        dispWHRE = re.compile('\s*DisplayWidth=(?P<width>\d+) *DisplayHeight=(?P<height>\d+)')
        out = self.raw_shell('dumpsys window')
        m = phyDispRE.search(out, 0)
        if not m:
            m = dispWHRE.search(out, 0)
        if m:
            for prop in ['width', 'height']:
                displayInfo[prop] = int(m.group(prop))
            for prop in ['density']:
                d = self._getDisplayDensity(strip=True)
                if d:
                    displayInfo[prop] = d
                else:
                    # No available density information
                    displayInfo[prop] = -1.0
            return displayInfo

        # gets C{mPhysicalDisplayInfo} values from dumpsys. This is a method to obtain display dimensions and density
        phyDispRE = re.compile('Physical size: (?P<width>\d+)x(?P<height>\d+).*Physical density: (?P<density>\d+)', re.S)
        m = phyDispRE.search(self.cmd('shell wm size; wm density', timeout=3))
        if m:
            for prop in ['width', 'height']:
                displayInfo[prop] = int(m.group(prop))
            for prop in ['density']:
                displayInfo[prop] = float(m.group(prop))
            return displayInfo

        if not displayInfo:
            raise DeviceConnectionError("Getting device screen information timed out")

    def _getDisplayDensity(self, strip=True):
        """
        Get display density

        Args:
            strip: strip the output

        Returns:
            display density

        """
        BASE_DPI = 160.0
        d = self.getprop('ro.sf.lcd_density', strip)
        if d:
            return float(d) / BASE_DPI
        d = self.getprop('qemu.sf.lcd_density', strip)
        if d:
            return float(d) / BASE_DPI
        return -1.0

    def getDisplayOrientation(self):
        """
        Another way to get the display orientation, this works well for older devices (SDK version 15)

        Returns:
            display orientation information

        """
        # another way to get orientation, for old sumsung device(sdk version 15) from xiaoma
        SurfaceFlingerRE = re.compile('orientation=(\d+)')
        output = self.shell('dumpsys SurfaceFlinger')
        m = SurfaceFlingerRE.search(output)
        if m:
            return int(m.group(1))

        # Fallback method to obtain the orientation
        # See https://github.com/dtmilano/AndroidViewClient/issues/128
        surfaceOrientationRE = re.compile('SurfaceOrientation:\s+(\d+)')
        output = self.shell('dumpsys input')
        m = surfaceOrientationRE.search(output)
        if m:
            return int(m.group(1))

        displayFramesRE = re.compile(r"DisplayFrames.*r=(\d+)")
        output = self.shell('dumpsys window displays')
        m = displayFramesRE.search(output)
        if m:
            return int(m.group(1))

        # We couldn't obtain the orientation
        warnings.warn("Could not obtain the orientation, return 0")
        return 0

    def update_cur_display(self, display_info):
        """
        Some phones support resolution modification, try to get the modified resolution from dumpsys
        adb shell dumpsys window displays | find "cur="

        本方法虽然可以更好地获取到部分修改过分辨率的手机信息
        但是会因为cur=(\d+)x(\d+)的数值在不同设备上width和height的顺序可能不同，导致横竖屏识别出现问题
        airtest不再使用本方法作为通用的屏幕尺寸获取方法，但依然可用于部分设备获取当前被修改过的分辨率

        Examples:

            >>> # 部分三星和华为设备，若分辨率没有指定为最高，可能会导致点击偏移，可以用这个方式强制修改：
            >>> # For some Samsung and Huawei devices, if the resolution is not specified as the highest,
            >>> # it may cause click offset, which can be modified in this way:
            >>> dev = device()
            >>> info = dev.display_info
            >>> info2 = dev.adb.update_cur_display(info)
            >>> dev.display_info.update(info2)

        Args:
            display_info: the return of self.getPhysicalDisplayInfo()

        Returns:
            display_info

        """
        # adb shell dumpsys window displays | find "init="
        # 上面的命令行在dumpsys window里查找init=widthxheight，得到的结果是物理分辨率，且部分型号手机不止一个结果
        # 如果改为读取 cur=widthxheight 的数据，得到的是修改过分辨率手机的结果（例如三星S8）
        actual = self.shell("dumpsys window displays")
        arr = re.findall(r'cur=(\d+)x(\d+)', actual)
        if len(arr) > 0:
            # 强制设定宽度width为更小的数字、height为更大的数字，避免因为各手机厂商返回结果的顺序不同导致问题
            # Set the width to a smaller number and the height to a larger number
            width, height = min(list(map(int, arr[0]))), max(list(map(int, arr[0])))
            display_info['physical_width'] = display_info['width']
            display_info['physical_height'] = display_info['height']
            display_info['width'], display_info['height'] = width, height
        return display_info

    def get_top_activity(self):
        """
        Perform `adb shell dumpsys activity top` command search for the top activity

        Raises:
            AirtestError: if top activity cannot be obtained

        Returns:
            top activity as a tuple: (package_name, activity_name, pid)

        """
        dat = self.shell('dumpsys activity top')
        activityRE = re.compile(r'\s*ACTIVITY ([A-Za-z0-9_.$]+)/([A-Za-z0-9_.$]+) \w+ pid=(\d+)')
        # in Android8.0 or higher, the result may be more than one
        m = activityRE.findall(dat)
        if m:
            return m[-1]
        else:
            raise AirtestError("Can not get top activity, output:%s" % dat)

    def is_keyboard_shown(self):
        """
        Perform `adb shell dumpsys input_method` command and search for information if keyboard is shown

        Returns:
            True or False whether the keyboard is shown or not

        """
        dim = self.shell('dumpsys input_method')
        if dim:
            return "mInputShown=true" in dim
        return False

    def is_screenon(self):
        """
        Perform `adb shell dumpsys window policy` command and search for information if screen is turned on or off

        Raises:
            AirtestError: if screen state can't be detected

        Returns:
            True or False whether the screen is turned on or off

        """
        screenOnRE = re.compile('mScreenOnFully=(true|false)')
        m = screenOnRE.search(self.shell('dumpsys window policy'))
        if m:
            return m.group(1) == 'true'
        else:
            # MIUI11
            screenOnRE = re.compile('screenState=(SCREEN_STATE_ON|SCREEN_STATE_OFF)')
            m = screenOnRE.search(self.shell('dumpsys window policy'))
            if m:
                return m.group(1) == 'SCREEN_STATE_ON'
        raise AirtestError("Couldn't determine screen ON state")

    @retries(max_tries=3)
    def is_locked(self):
        """
        Perform `adb shell dumpsys window policy` command and search for information if screen is locked or not

        Raises:
            AirtestError: if lock screen can't be detected

        Returns:
            True or False whether the screen is locked or not

        """
        lockScreenRE = re.compile('(?:mShowingLockscreen|isStatusBarKeyguard|showing)=(true|false)')
        m = lockScreenRE.search(self.shell('dumpsys window policy'))
        if not m:
            raise AirtestError("Couldn't determine screen lock state")
        return (m.group(1) == 'true')

    def unlock(self):
        """
        Perform `adb shell input keyevent MENU` and `adb shell input keyevent BACK` commands to attempt
        to unlock the screen

        Returns:
            None

        Warnings:
            Might not work on all devices

        """
        self.shell('input keyevent MENU')
        self.shell('input keyevent BACK')

    def get_package_version(self, package):
        """
        Perform `adb shell dumpsys package` and search for information about given package version

        Args:
            package: package name

        Returns:
            None if no info has been found, otherwise package version

        """
        package_info = self.shell(['dumpsys', 'package', package])
        matcher = re.search(r'versionCode=(\d+)', package_info)
        if matcher:
            return int(matcher.group(1))
        return None

    def list_app(self, third_only=False):
        """
        Perform `adb shell pm list packages` to print all packages, optionally only
          those whose package name contains the text in FILTER.

        Options
            -f: see their associated file
            -d: filter to only show disabled packages
            -e: filter to only show enabled packages
            -s: filter to only show system packages
            -3: filter to only show third party packages
            -i: see the installer for the packages
            -u: also include uninstalled packages


        Args:
            third_only: print only third party packages

        Returns:
            list of packages

        """
        cmd = ["pm", "list", "packages"]
        if third_only:
            cmd.append("-3")
        output = self.shell(cmd)
        packages = output.splitlines()
        # remove all empty string; "package:xxx" -> "xxx"
        packages = [p.split(":")[1] for p in packages if p]
        return packages

    def path_app(self, package):
        """
        Perform `adb shell pm path` command to print the path to the package

        Args:
            package: package name

        Raises:
            AdbShellError: if any adb error occurs
            AirtestError: if package is not found on the device

        Returns:
            path to the package

        """
        try:
            output = self.shell(['pm', 'path', package])
        except AdbShellError:
            output = ""
        if 'package:' not in output:
            raise AirtestError('package not found, output:[%s]' % output)
        return output.split("package:")[1].strip()

    def check_app(self, package):
        """
        Perform `adb shell dumpsys package` command and check if package exists on the device

        Args:
            package: package name

        Raises:
            AirtestError: if package is not found

        Returns:
            True if package has been found

        """
        output = self.shell(['dumpsys', 'package', package])
        pattern = r'Package\s+\[' + str(package) + '\]'
        match = re.search(pattern, output)
        if match is None:
            raise AirtestError('package "{}" not found'.format(package))
        return True

    def start_app(self, package, activity=None):
        """
        Perform `adb shell monkey` commands to start the application, if `activity` argument is `None`, then
        `adb shell am start` command is used.

        Args:
            package: package name
            activity: activity name

        Returns:
            None

        """
        if not activity:
            try:
                ret = self.shell(['monkey', '-p', package, '-c', 'android.intent.category.LAUNCHER', '1'])
            except AdbShellError as e:
                raise AirtestError("Starting App: %s Failed! No activities found to run." % package)
            if "No activities found to run" in ret:
                raise AirtestError("Starting App: %s Failed! No activities found to run." % package)
        else:
            self.shell(['am', 'start', '-n', '%s/%s.%s' % (package, package, activity)])

    def start_app_timing(self, package, activity):
        """
        Start the application and activity, and measure time

        Args:
            package: package name
            activity: activity name

        Returns:
            app launch time

        """
        out = self.shell(['am', 'start', '-S', '-W', '%s/%s' % (package, activity),
                          '-c', 'android.intent.category.LAUNCHER', '-a', 'android.intent.action.MAIN'])
        if not re.search(r"Status:\s*ok", out):
            raise AirtestError("Starting App: %s/%s Failed!" % (package, activity))

        # matcher = re.search(r"TotalTime:\s*(\d+)", out)
        matcher = re.search(r"TotalTime:\s*(\d+)", out)
        if matcher:
            return int(matcher.group(1))
        else:
            return 0

    def stop_app(self, package):
        """
        Perform `adb shell am force-stop` command to force stop the application

        Args:
            package: package name

        Returns:
            None

        """
        self.shell(['am', 'force-stop', package])

    def clear_app(self, package):
        """
        Perform `adb shell pm clear` command to clear all application data

        Args:
            package: package name

        Returns:
            None

        """
        self.shell(['pm', 'clear', package])

    def text(self, content):
        """
        Use adb shell input for text input

        Args:
            content: text to input

        Returns:
            None

        Examples:
            >>> dev = connect_device("Android:///")
            >>> dev.text("Hello World")
            >>> dev.text("test123")
        """
        if content.isalpha():
            self.shell(["input", "text", content])
        else:
            # If it contains letters and numbers, input with `adb shell input text` may result in the wrong order
            for i in content:
                if i == " ":
                    self.keyevent("KEYCODE_SPACE")
                else:
                    self.shell(["input", "text", i])

    def get_ip_address(self):
        """
        Perform several set of commands to obtain the IP address.

            * `adb shell netcfg | grep wlan0`
            * `adb shell ifconfig`
            * `adb getprop dhcp.wlan0.ipaddress`

        Returns:
            None if no IP address has been found, otherwise return the IP address

        """

        def get_ip_address_from_interface(interface):
            """Get device ip from target network interface."""
            # android >= 6.0: ip -f inet addr show {interface}
            try:
                res = self.shell('ip -f inet addr show {}'.format(interface))
            except AdbShellError:
                res = ''
            matcher = re.search(r"inet (\d+\.){3}\d+", res)
            if matcher:
                return matcher.group().split(" ")[-1]

            # android >= 6.0 backup method: ifconfig
            try:
                res = self.shell('ifconfig')
            except AdbShellError:
                res = ''
            matcher = re.search(interface + r'.*?inet addr:((\d+\.){3}\d+)', res, re.DOTALL)
            if matcher:
                return matcher.group(1)

            # android <= 6.0: netcfg
            try:
                res = self.shell('netcfg')
            except AdbShellError:
                res = ''
            matcher = re.search(interface + r'.* ((\d+\.){3}\d+)/\d+', res)
            if matcher:
                return matcher.group(1)

            # android <= 6.0 backup method: getprop dhcp.{}.ipaddress
            try:
                res = self.shell('getprop dhcp.{}.ipaddress'.format(interface))
            except AdbShellError:
                res = ''
            matcher = IP_PATTERN.search(res)
            if matcher:
                return matcher.group(0)

            # sorry, no more methods...
            return None

        interfaces = ('eth0', 'eth1', 'wlan0')
        for i in interfaces:
            ip = get_ip_address_from_interface(i)
            if ip and not ip.startswith('127.') and not ip.startswith('169.'):
                return ip

        return None

    def get_gateway_address(self):
        """
        Perform several set of commands to obtain the gateway address.
            * `adb getprop dhcp.wlan0.gateway`
            * `adb shell netcfg | grep wlan0`

        Returns:
            None if no gateway address has been found, otherwise return the gateway address

        """
        ip2int = lambda ip: reduce(lambda a, b: (a << 8) + b, map(int, ip.split('.')), 0)
        int2ip = lambda n: '.'.join([str(n >> (i << 3) & 0xFF) for i in range(0, 4)[::-1]])
        try:
            res = self.shell('getprop dhcp.wlan0.gateway')
        except AdbShellError:
            res = ''
        matcher = IP_PATTERN.search(res)
        if matcher:
            return matcher.group(0)
        ip = self.get_ip_address()
        if not ip:
            return None
        mask_len = self._get_subnet_mask_len()
        gateway = (ip2int(ip) & (((1 << mask_len) - 1) << (32 - mask_len))) + 1
        return int2ip(gateway)

    def _get_subnet_mask_len(self):
        """
        Perform `adb shell netcfg | grep wlan0` command to obtain mask length

        Returns:
            17 if mask length could not be detected, otherwise the mask length

        """
        try:
            res = self.shell('netcfg')
        except AdbShellError:
            pass
        else:
            matcher = re.search(r'wlan0.* (\d+\.){3}\d+/(\d+) ', res)
            if matcher:
                return int(matcher.group(2))
        # 获取不到网段长度就默认取17
        print('[iputils WARNING] fail to get subnet mask len. use 17 as default.')
        return 17

    def get_memory(self):
        res = self.shell("dumpsys meminfo")
        pat = re.compile(r".*Total RAM:\s+(\S+)\s+", re.DOTALL)
        _str = pat.match(res).group(1)
        if ',' in _str:
            _list = _str.split(',')
            _num = int(_list[0])
            _num = round(_num + (float(_list[1]) / 1000.0))
        else:
            _num = round(float(_str) / 1000.0 / 1000.0)
        res = str(_num) + 'G'
        return res

    def get_storage(self):
        res = self.shell("df /data")
        pat = re.compile(r".*\/data\s+(\S+)", re.DOTALL)
        if pat.match(res):
            _str = pat.match(res).group(1)
        else:
            pat = re.compile(r".*\s+(\S+)\s+\S+\s+\S+\s+\S+\s+\/data", re.DOTALL)
            _str = pat.match(res).group(1)
        if 'G' in _str:
            _num = round(float(_str[:-1]))
        elif 'M' in _str:
            _num = round(float(_str[:-1]) / 1000.0)
        else:
            _num = round(float(_str) / 1000.0 / 1000.0)
        if _num > 64:
            res = '128G'
        elif _num > 32:
            res = '64G'
        elif _num > 16:
            res = '32G'
        elif _num > 8:
            res = '16G'
        else:
            res = '8G'
        return res

    def get_cpuinfo(self):
        res = self.shell("cat /proc/cpuinfo").strip()
        cpuNum = res.count("processor")
        pat = re.compile(r'Hardware\s+:\s+(\w+.*)')
        m = pat.search(res)
        if not m:
            pat = re.compile(r'Processor\s+:\s+(\w+.*)')
            m = pat.match(res)
        cpuName = m.group(1).replace('\r', '')
        return dict(cpuNum=cpuNum, cpuName=cpuName)

    def get_cpufreq(self):
        res = self.shell("cat /sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_max_freq")
        num = round(float(res) / 1000 / 1000, 1)
        res = str(num) + 'GHz'
        return res.strip()

    def get_cpuabi(self):
        res = self.shell("getprop ro.product.cpu.abi")
        return res.strip()

    def get_gpu(self):
        res = self.shell("dumpsys SurfaceFlinger")
        pat = re.compile(r'GLES:\s+(.*)')
        m = pat.search(res)
        if not m:
            return None
        _list = m.group(1).split(',')
        gpuModel = ""
        opengl = ""
        if len(_list) > 0:
            gpuModel = _list[1].strip()
        if len(_list) > 1:
            m2 = re.search(r'(\S+\s+\S+\s+\S+).*', _list[2])
            if m2:
                opengl = m2.group(1)
        return dict(gpuModel=gpuModel, opengl=opengl)

    def get_model(self):
        return self.getprop("ro.product.model")

    def get_manufacturer(self):
        return self.getprop("ro.product.manufacturer")

    def get_device_info(self):
        """
        Get android device information, including: memory/storage/display/cpu/gpu/model/manufacturer...

        Returns:
            Dict of info

        """
        handlers = {
            "platform": "Android",
            "serialno": self.serialno,
            "memory": self.get_memory,
            "storage": self.get_storage,
            "display": self.getPhysicalDisplayInfo,
            "cpuinfo": self.get_cpuinfo,
            "cpufreq": self.get_cpufreq,
            "cpuabi": self.get_cpuabi,
            "sdkversion": self.sdk_version,
            "gpu": self.get_gpu,
            "model": self.get_model,
            "manufacturer": self.get_manufacturer,
            # "battery": getBatteryCapacity
        }
        ret = {}
        for k, v in handlers.items():
            if callable(v):
                try:
                    value = v()
                except Exception:
                    value = None
                ret[k] = value
            else:
                ret[k] = v
        return ret

    def get_display_of_all_screen(self, info, package=None):
        """
        Perform `adb shell dumpsys window windows` commands to get window display of application.

        Args:
            info: device screen properties
            package: package name, default to the package of top activity

        Returns:
            None if adb command failed to run, otherwise return device screen properties(portrait mode)
            eg. (offset_x, offset_y, screen_width, screen_height)

        """
        output = self.shell("dumpsys window windows")
        windows = output.split("Window #")
        offsetx, offsety, width, height = 0, 0, info['width'], info['height']
        package = self._search_for_current_package(output) if package is None else package
        if package:
            for w in windows:
                if "package=%s" % package in w:
                    arr = re.findall(r'Frames: containing=\[(\d+\.?\d*),(\d+\.?\d*)]\[(\d+\.?\d*),(\d+\.?\d*)]', w)
                    if len(arr) >= 1 and len(arr[0]) == 4:
                        offsetx, offsety, width, height = float(arr[0][0]), float(arr[0][1]), float(arr[0][2]), float(arr[0][3])
                        if info["orientation"] in [1, 3]:
                            offsetx, offsety, width, height = offsety, offsetx, height, width
                        width, height = width - offsetx, height - offsety
        return {
            "offset_x": offsetx,
            "offset_y": offsety,
            "offset_width": width,
            "offset_height": height,
        }

    def _search_for_current_package(self, ret):
        """
        Search for current app package name from the output of command "adb shell dumpsys window windows"

        Returns:
            package name if exists else ""
        """
        try:
            packageRE = re.compile('\s*mCurrentFocus=Window{.* ([A-Za-z0-9_.]+)/[A-Za-z0-9_.]+}')
            m = packageRE.findall(ret)
            if m:
                return m[-1]
            else:
                return self.get_top_activity()[0]
        except Exception as e:
            print("[Error] Cannot get current top activity")
        return ""


def cleanup_adb_forward():
    for adb in ADB._instances:
        adb._cleanup_forwards()


reg_cleanup(cleanup_adb_forward)



import os
import time
from collections import deque

from airtestProject.airtest.core import api as air
from airtestProject.airtest.core.android.recorder import Recorder
from airtestProject.airtest.report.report import LogToHtml

from airtestProject.commons.utils.logger import log
from airtestProject.factory.operateFactory import operate


class FuncListCheckPage:

    def __init__(self, adb, script_root, Project=None, log_path=None):
        self.adb = adb
        if Project is not None:
            operate('air').set_dict(script_root, Project)
        if log_path is not None:
            self.log_path = log_path
        else:
            self.log_path = None

    def open_ui_from_ui_list(self, ui, ui_view, swipe_point, first_ui, final_ui, fun_name="air"):
        """

        :param ui: 需要点击的ui
        :param ui_view: 打开后的ui界面
        :param swipe_point: 滑动点
        :param first_ui: 列表第一元素
        :param final_ui: 列表最后元素
        :param fun_name: 选用poco或air
        :return:
        """

        @log.wrap(f'正在打开{ui}界面')
        def _open_ui_from_ui_list():
            self.adb.shell("logcat -c")
            while operate(fun_name).exists(ui) is False:
                operate(fun_name).swipe(swipe_point, vector_direction=[0, -0.2])
                if operate(fun_name).exists(final_ui) is not False:
                    start_time = time.time()
                    while operate(fun_name).exists(ui) is False:
                        if time.time() - start_time > 300 or operate(fun_name).exists(first_ui) is not False:
                            log.info(f"无法找到该{ui}请检查调整参数或查看页面")
                            return
                        operate(fun_name).swipe(swipe_point, vector_direction=[0, 0.18])
            if operate(fun_name).wait_next_element(ui, ui_view) is True:
                air.sleep(3)
                log.info(f"成功点击打开{ui}")
                adb_log = self.adb.shell("logcat -d Unity:E *:S")
                log.log_adb_out(adb_log)
            else:
                air.sleep(3)
                log.info(f"打开{ui}失败")
                adb_log = self.adb.shell("logcat -d Unity:W *:S")
                log.log_adb_out(adb_log)

        _open_ui_from_ui_list()

    def open_ui(self, ui, ui_view, fun_name="air"):

        @log.wrap(f'正在打开{ui}界面')
        def _open_ui():
            self.adb.shell("logcat -c")
            if operate(fun_name).wait_next_element(ui, ui_view) is True:
                air.sleep(5)
                log.info(f"成功点击打开{ui}")
                adb_log = self.adb.shell("logcat -d Unity:E *:S")
                log.log_adb_out(adb_log)
            else:
                air.sleep(5)
                log.info(f"打开{ui}失败")
                adb_log = self.adb.shell("logcat -d Unity:W *:S")
                log.log_adb_out(adb_log)

        _open_ui()

    def close_ui(self, ui_view, close_pos, fun_name='air'):
        """

        :param ui_view: 当前界面
        :param close_pos: 关闭按钮
        :param fun_name: 选用poco或air
        :return:
        """
        self.adb.shell("logcat -c")

        @log.wrap(f'关闭ui界面{ui_view}')
        def _close_ui():
            if operate(fun_name).wait_element_appear(close_pos):
                operate(fun_name).click(close_pos)
                if operate(fun_name).wait_disappear_element(ui_view):
                    adb_out = self.adb.shell("logcat -d Unity:E *:S")
                    log.log_adb_out(adb_out)
                    log.info("关闭ui界面")
                else:
                    log.error(f"关闭{ui_view}失败")
                    adb_out = self.adb.shell("logcat -d Unity:W *:S")
                    log.log_adb_out(adb_out)

        _close_ui()

    def click_ui(self, ui, ocrPlus=None, fun_name='air'):

        @log.wrap(f"点击{ui}")
        def _click_ui():
            self.adb.shell("logcat -c")
            if ocrPlus:
                operate(fun_name).click(ui, ocrPlus=ocrPlus)
            operate(fun_name).click(ui)
            air.sleep(2)
            adb_out = self.adb.shell("logcat -d Unity:W *:S")
            log.log_adb_out(adb_out)
            log.info("点击ui")

        _click_ui()

    def click_ui_from_ui_list(self, ui, swipe_point, first_ui, final_ui, fun_name='air'):
        self.adb.shell("logcat -c")

        @log.wrap(f"从列表点击{ui}")
        def _click_ui_from_ui_list():
            while operate(fun_name).exists(ui) is False:
                operate(fun_name).swipe(swipe_point, vector_direction=[0, -0.2])
                if operate(fun_name).exists(final_ui) is not False:
                    start_time = time.time()
                    while operate(fun_name).exists(ui) is False:
                        if time.time() - start_time > 300 or operate(fun_name).exists(first_ui) is not False:
                            log.info(f"无法找到该{ui}请检查调整参数或查看页面")
                            return
                        operate(fun_name).swipe(swipe_point, vector_direction=[0, 0.1])
            operate(fun_name).click(ui)
            air.sleep(3)
            adb_out = self.adb.shell("logcat -d Unity:W *:S")
            log.log_adb_out(adb_out)
            log.info("点击ui")

        _click_ui_from_ui_list()

    def open_list(self, list_pos, next_pos, fun_name="air"):
        @log.wrap(f'打开{list_pos}列表')
        def _open_list():
            if operate(fun_name).wait_next_element(list_pos, next_pos) is False:
                log.error('打开列表失败')

        _open_list()

    def check_scenes(self, scenes_pos, wait_pos, fun_name="air"):
        self.adb.shell("logcat -c")

        @log.wrap(f'跳转场景{scenes_pos}')
        def _check_scenes():
            operate(fun_name).click(scenes_pos)
            air.sleep(0.2)
            if operate(fun_name).exists(wait_pos):
                operate(fun_name).wait_disappear_element(wait_pos)
                air.sleep(0.5)
                adb_out = self.adb.shell("logcat -d Unity:W *:S")
                log.log_adb_out(adb_out)
            else:
                log.info("点击无反应，可能是已处于对应场景")
                # air.snapshot(filename="screenshot{}点击无反应的界面.png".format(scenes_pos))
                air.sleep(0.5)
                adb_out = self.adb.shell("logcat -d Unity:W *:S")
                log.log_adb_out(adb_out)
                print(adb_out)
                return False

        _check_scenes()

    def swip_ui(self, swipe_pos, fun_name="air"):

        @log.wrap(f"滑动测试{swipe_pos}")
        def _swip_ui():
            self.adb.shell("logcat -c")
            operate(fun_name).swipe(swipe_pos)
            air.sleep(2)
            adb_out = self.adb.shell("logcat -d Unity:W *:S")
            log.log_adb_out(adb_out)
            log.info("滑动界面")

        _swip_ui()

    # TODO: 队列版本后续开发一个识别黑框的会方便点（poco其实不用这么麻烦，我后续想想怎么整合(poco可以知道ui节点也就知道了黑框存在)）
    def Plot(self, dialogue_ui_list, stop_ui, choice_ui=None, jump_ui=None, start_ui=None, log_path=None,
             fun_name="air"):
        """

        :param dialogue_ui_list: 对话list（记得按顺序，以队列遍历，遍历一个删一个）
        :param stop_ui: 剧情停止标识
        :param choice_ui: 选择按钮列表
        :param jump_ui: 跳过按钮可填无
        :param start_ui: 开始剧情界面
        :param log_path: 视频保存路径
        :param fun_name: poco或airtest
        :return:
        """
        self.adb.shell("logcat -c")
        recorder = Recorder(self.adb)
        if choice_ui is None:
            choice_ui = []
        recorder.start_recording()

        @log.wrap("过剧情")
        def _Plot():
            if start_ui is not None:
                operate(fun_name).wait_element_appear(start_ui)
            dialogue_ui_queue = deque(dialogue_ui_list)
            while operate(fun_name).exists(stop_ui) is False:
                temp = True
                if jump_ui and operate(fun_name).exists(jump_ui):
                    operate(fun_name).click(jump_ui)
                    break
                for i in choice_ui:
                    if operate(fun_name).exists(i):
                        operate(fun_name).click(i)
                        temp = False
                        break
                if temp is True:
                    while dialogue_ui_queue:
                        dialogue = dialogue_ui_queue.popleft()
                        if operate(fun_name).exists(dialogue):
                            operate(fun_name).click(dialogue)
                            break
            air.sleep(2)
            adb_out = self.adb.shell("logcat -d Unity:W *:S")
            log.log_adb_out(adb_out)
            log.info("过完剧情")

        _Plot()
        if log_path is not None:
            recorder.stop_recording(output=log_path + r"\剧情视频{}.mp4".format(start_ui))
        else:
            recorder.stop_recording()

    def scenes_jump(self):
        self.open_list("Scene", "场景跳转")
        self.check_scenes("斗罗场景", "WaitingBar")
        # self.open_list("Scene", "场景跳转")
        # self.check_scenes("TestShader", "WaitingBar")
        # self.open_list("Scene", "场景跳转")
        # self.check_scenes("斗罗场景 (流式加载)", "WaitingBar")

    def she_zhi_view(self):
        self.open_list("功能", "功能列表")
        self.open_ui_from_ui_list("设置", "SettingsPage", "功能列表滑动点", "LanguageRoom", "Multilingual")
        self.close_ui("SettingsPage", "CloseSet")

    def hua_zhi_view(self):
        self.open_list("Func", "功能列表")
        self.open_ui_from_ui_list("画质", "PictureQualityPage", "FuncListSwipe", "LanguageRoom", "Multilingual")
        self.click_ui("veryLow")
        self.click_ui("Application")
        self.click_ui("Low")
        self.click_ui("Application")
        self.click_ui("Mid")
        self.click_ui("Application")
        self.click_ui("其它")
        self.click_ui("画质")
        self.close_ui("PictureQualityPage", "BackButton")

    def ji_neng_view(self):
        self.open_list("Func", "功能列表")
        self.open_ui_from_ui_list("技能装配UI", "skillView", "FuncListSwipe", "LanguageRoom", "Multilingual")
        self.click_ui("skillViewAttackKey")
        self.close_ui("skillView", "BackButton")

    def demo_ui_view(self):
        self.open_list("Func", "功能列表")
        self.open_ui_from_ui_list("DemoUI", "demoUIView", "FuncListSwipe", "LanguageRoom", "Multilingual")
        self.click_ui("点击发事件")
        self.click_ui("列表组件")
        self.click_ui("基础组件")
        self.close_ui("demoUIView", "BackButton")

    def sound_setting_view(self):
        self.open_list("Func", "功能列表")
        self.open_ui_from_ui_list("soudsetingview", "MusicSettingInterface", "FuncListSwipe", "LanguageRoom", "Multilingual")
        self.click_ui("使用Wwise", ocrPlus=True)
        self.click_ui("Destroy Audio")
        self.click_ui("使用 UnityAudio")
        self.click_ui("创建机关")
        self.close_ui("MusicSettingInterface", "CloseMusicView")

    def DemoRTShowHighQualityShadowView(self):
        self.open_list("Func", "功能列表")
        self.open_ui_from_ui_list("DemoRTShowHighQualityShadowView", "HDshadowModelInterface", "FuncListSwipe", "LanguageRoom",
                                  "Multilingual")
        self.close_ui("HDshadowModelInterface", "BackButton")

    def DemoRTShowOrthographicView(self):
        self.open_list("Func", "功能列表")
        self.open_ui_from_ui_list("DemoRTShowOrthographicView", "DemoRTShowOrthographicViewIn", "FuncListSwipe", "LanguageRoom",
                                  "Multilingual")
        self.close_ui("DemoRTShowOrthographicViewIn", "BackButton")

    def DemoRTShowPlanarShadowView(self):
        self.open_list("Func", "功能列表")
        self.open_ui_from_ui_list("DemoRTShowPlanarShadowView", "PlanarShadow", "FuncListSwipe", "LanguageRoom",
                                  "Multilingual")
        self.close_ui("PlanarShadow", "BackButton")

    def SRPBatcherProfiler(self):
        self.open_list("Func", "功能列表")
        self.open_ui_from_ui_list("SRPBatcherProfiler", "关闭SRP Batcher", "FuncListSwipe", "LanguageRoom",
                                  "Multilingual")
        self.click_ui("关闭SRP Batcher")
        self.click_ui("开启SRP Batcher")
        self.open_list("Func", "功能列表")
        self.close_ui("关闭SRP Batcher", "SRPBatcherProfiler")

    def TestUIMoldeClipView(self):
        self.open_list("Func", "功能列表")
        self.open_ui_from_ui_list("TestUIMoldeClipView", "TestMoldeView", "FuncListSwipe", "LanguageRoom",
                                  "Multilingual")
        self.swip_ui("TestMoldeSwipe")
        self.click_ui("打开新")
        self.click_ui("TestUIMoldeClipView")
        self.close_ui("TestMoldeView", "TestMoldeCloseBtn")

    def xie_yi_view(self):
        self.open_list("Func", "功能列表")
        self.open_ui_from_ui_list("Proto", "测试发送", "FuncListSwipe", "LanguageRoom", "Multilingual")
        self.click_ui("恢复监听发送")
        self.click_ui("恢复监听接收")
        self.click_ui("测试发送")
        self.click_ui("清空")
        self.close_ui("测试发送", "ProtoCloseBtn")

    def shi_pin(self):
        self.open_list("Func", "功能列表")
        self.open_ui_from_ui_list("视频", "VideoView", "FuncListSwipe", "LanguageRoom", "Multilingual")
        self.close_ui("VideoView", "VideoInterfaceCloseButton")

    def ce_shi_ju_qin1(self, log_path):
        self.open_list("Func", "功能列表")
        self.open_ui_from_ui_list("测试剧情1", "TestPlotBlackCurtain1", "FuncListSwipe", "LanguageRoom", "Multilingual")
        self.Plot(["最近怪物横行，请少侠前往击杀"], "功能", ["挑战Boss"], start_ui="TestPlotStart1", log_path=log_path)

    def ce_shi_ju_qin2(self, log_path):
        self.open_list("Func", "功能列表")
        self.open_ui_from_ui_list("测试剧情2", "TestPlotBlackCurtain2", "FuncListSwipe", "LanguageRoom", "Multilingual")
        self.Plot(["最近怪物横行，请少侠前往击杀", "这个小怪有点强，你要不要和我一起去？", "不是吧，小怪还要我帮忙啊！",
                  "两个人一起，不是打的快些嘛", "好吧，陪你走一遭"], stop_ui="功能", start_ui="TestPlotStart2", log_path=log_path)

    def run_step(self):
        self.scenes_jump()

        # self.she_zhi_view()
        #
        # self.hua_zhi_view()
        #
        # self.open_list("Func", "功能列表")
        # self.click_ui_from_ui_list("reloadall", "FuncListSwipe", "LanguageRoom", "Multilingual")
        #
        # self.ji_neng_view()
        #
        # self.open_list("Func", "功能列表")
        # self.click_ui_from_ui_list("testloadrawdata", "FuncListSwipe", "LanguageRoom", "Multilingual")
        #
        # self.demo_ui_view()
        #
        # self.sound_setting_view()
        #
        # self.DemoRTShowHighQualityShadowView()
        #
        # self.DemoRTShowOrthographicView()
        #
        # self.DemoRTShowPlanarShadowView()
        #
        # self.SRPBatcherProfiler()
        #
        # self.TestUIMoldeClipView()
        #
        # self.open_list("Func", "功能列表")
        # self.click_ui_from_ui_list("checkForUpdates", "FuncListSwipe", "LanguageRoom", "Multilingual")
        # self.open_list("Func", "功能列表")
        # self.click_ui_from_ui_list("镜头切换", "FuncListSwipe", "LanguageRoom", "Multilingual")
        # self.click_ui_from_ui_list("镜头切换", "FuncListSwipe", "LanguageRoom", "Multilingual")
        # self.click_ui_from_ui_list("镜头切换", "FuncListSwipe", "LanguageRoom", "Multilingual")
        #
        # self.xie_yi_view()
        #
        # self.shi_pin()
        #
        # self.open_list("Func", "功能列表")
        # self.click_ui_from_ui_list("whetherOrNotObjID", "FuncListSwipe", "LanguageRoom", "Multilingual")
        # self.click_ui_from_ui_list("whetherOrNotObjID", "FuncListSwipe", "LanguageRoom", "Multilingual")
        # self.click_ui_from_ui_list("记录协议", "FuncListSwipe", "LanguageRoom", "Multilingual")
        # self.click_ui_from_ui_list("显示ping", "FuncListSwipe", "LanguageRoom", "Multilingual")
        # self.click_ui_from_ui_list("errorPromptView", "FuncListSwipe", "LanguageRoom", "Multilingual")
        # self.click_ui_from_ui_list("DamageText Test", "FuncListSwipe", "LanguageRoom", "Multilingual")
        # self.open_list("Func", "功能列表")
        # self.click_ui_from_ui_list("GCTest", "FuncListSwipe", "LanguageRoom", "Multilingual")
        # self.open_list("Func", "功能列表")
        # self.click_ui_from_ui_list("测试报错", "FuncListSwipe", "LanguageRoom", "Multilingual")
        #
        # self.ce_shi_ju_qin1(self.log_path)
        #
        # self.ce_shi_ju_qin2(self.log_path)


def _wait_for_log_generation(log_path, timeout=60):
    size = -1
    start_time = time.time()
    while time.time() - start_time < timeout:
        if not os.path.exists(log_path):
            time.sleep(1)
            continue
        current_size = os.path.getsize(log_path)
        if current_size == size:
            break  # 文件大小没有变化，认为日志生成完毕
        else:
            size = current_size
            time.sleep(1)  # 等待一秒再次检查


if __name__ == '__main__':
    project_root = r"G:\pyProject\odin-testautomation\TestAutomation\airtestProjects\commons\page\odin"
    log_path = r"/airtestProject\commons\page\odin\log"
    air.auto_setup(__file__, devices=["Android:///"], project_root=project_root, logdir=True)
    serialno = air.device().adb.serialno
    # output_html_path = r"G:\pyProject\odin-testautomation\TestAutomation\airtestProject\commons\page\odin\log\log
    # .html"
    _wait_for_log_generation(log_path)
    func_list_check_report = LogToHtml(script_root=__file__, log_root=log_path,
                                       export_dir=project_root + r"\report", logfile=log_path + r"\log.txt", lang='zh',
                                       plugins=None)
    func_list_check_report.report()



# coding=utf-8

"""本文件用于存放一些计算函数，比如图像适配、搜索区域预测."""


def no_resize(w_a, h_a, resolution_a, resolution_b):
    """无缩放策略."""
    return w_a, h_a


def cocos_min_strategy(w, h, sch_resolution, src_resolution, design_resolution=(960, 640)):
    """图像缩放规则: COCOS中的MIN策略."""
    # 输入参数: w-h待缩放图像的宽高，sch_resolution为待缩放图像的来源分辨率
    #           src_resolution 待适配屏幕的分辨率  design_resolution 软件的设计分辨率
    # 需要分别算出对设计分辨率的缩放比，进而算出src\sch有效缩放比。
    scale_sch = min(1.0 * sch_resolution[0] / design_resolution[0], 1.0 * sch_resolution[1] / design_resolution[1])
    scale_src = min(1.0 * src_resolution[0] / design_resolution[0], 1.0 * src_resolution[1] / design_resolution[1])
    scale = scale_src / scale_sch
    h_re, w_re = int(h * scale), int(w * scale)
    return w_re, h_re


def predict_area(im_source, op_pos, radius_x, radius_y, src_resolution=None):
    """根据参数进行screen的预测区域."""
    # 预测操作位置: (按照比例进行点预测) clk_x, clk_y是规划为比例的
    clk_x, clk_y = op_pos
    # 如果没有传递
    if src_resolution:
        res_x, res_y = src_resolution
    else:
        res_y, res_x = im_source.shape[:2]
    prePos_x, prePos_y = clk_x * res_x + 0.5 * res_x, clk_y * res_x + 0.5 * res_y

    def safe_xy(val, min_val, max_val):
        return min(max(min_val, val), max_val)

    # 再以预测点为中心，进行范围提取:
    start_x = int(safe_xy(prePos_x - radius_x, 0, res_x - 1))
    end_x = int(safe_xy(prePos_x + radius_x, 0, res_x - 1))
    start_y = int(safe_xy(prePos_y - radius_y, 0, res_y - 1))
    end_y = int(safe_xy(prePos_y + radius_y, 0, res_y - 1))

    # 如果发现预测区域完全在图像外，预测区域将只剩下一条像素，预测失败，直接raise:
    if start_x == end_x or start_y == end_y:
        img_src, left_top_pos = im_source, (0, 0)
        log_info = "Predict area's width or height has just one pixel, abandon prediction."
    else:
        # 预测区域正常，则截取预测区域，并将预测区域在源图像中的位置一并返回:
        img_src = im_source[start_y:end_y, start_x:end_x]
        left_top_pos = (start_x, start_y)
        # 输出调试信息.
        log_info = "predict rect:  X (%(start_x)s:%(end_x)s)   Y (%(start_y)s:%(end_y)s)" % {"start_x": start_x, "end_x":end_x, "start_y": start_y, "end_y": end_y}

    return img_src, left_top_pos, log_info



# coding=utf-8

from airtestProject.poco.sdk.Attributor import Attributor
from airtestProject.poco.sdk.exceptions import UnableToSetAttributeException


class StdAttributor(Attributor):
    def __init__(self, client):
        super(StdAttributor, self).__init__()
        self.client = client

    def setAttr(self, node, attrName, attrVal):
        if attrName == 'text':
            if type(node) in (list, tuple):
                node = node[0]
            instance_id = node.getAttr('_instanceId')
            if instance_id:
                success = self.client.call('SetText', instance_id, attrVal)
                if success:
                    return True
        raise UnableToSetAttributeException(attrName, node)



# -*- coding: utf-8 -*-
# @Author: gzliuxin
# @Email:  gzliuxin@corp.netease.com
# @Date:   2017-07-14 19:47:51

from airtestProject.poco.pocofw import Poco
from airtestProject.poco.agent import PocoAgent
from airtestProject.poco.freezeui.hierarchy import FrozenUIHierarchy, FrozenUIDumper
from airtestProject.poco.utils.simplerpc.utils import sync_wrapper
from airtestProject.poco.utils.airtest import AirtestInput, AirtestScreen
from airtestProject.poco.utils.simplerpc.rpcclient import RpcClient
from airtestProject.poco.utils.simplerpc.transport.ws import WebSocketClient
from airtestProject.poco.utils import six
if six.PY3:
    from urllib.parse import urlparse
else:
    from urlparse import urlparse

from airtestProject.airtest.core.api import connect_device, device as current_device
from airtestProject.poco.utils.device import default_device
from airtestProject.airtest.core.helper import device_platform


__all__ = ['CocosJsPoco']
DEFAULT_PORT = 5003
DEFAULT_ADDR = ('localhost', DEFAULT_PORT)


class CocosJsPocoAgent(PocoAgent):
    def __init__(self, port, device=None, ip=None):
        if ip is None or ip == "localhost":
            self.device = device or default_device()

            platform_name = device_platform(self.device)
            if platform_name == 'Android':
                local_port, _ = self.device.adb.setup_forward('tcp:{}'.format(port))
                ip = self.device.adb.host or 'localhost'
                port = local_port
            elif platform_name == 'IOS':
                port, _ = self.device.setup_forward(port)
                if self.device.is_local_device:
                    ip = 'localhost'
                else:
                    ip = self.device.ip
            else:
                ip = self.device.get_ip_address()

        # transport
        self.conn = WebSocketClient('ws://{}:{}'.format(ip, port))
        self.c = RpcClient(self.conn)
        self.c.connect()

        hierarchy = FrozenUIHierarchy(Dumper(self.c))
        screen = AirtestScreen()
        inputs = AirtestInput()
        super(CocosJsPocoAgent, self).__init__(hierarchy, inputs, screen, None)

    @property
    def rpc(self):
        return self.c

    def get_sdk_version(self):
        return self.rpc.call("getSDKVersion")


class Dumper(FrozenUIDumper):
    def __init__(self, rpcclient):
        super(Dumper, self).__init__()
        self.rpcclient = rpcclient

    @sync_wrapper
    def dumpHierarchy(self, onlyVisibleNode=True):
        # NOTE: cocosjs 的driver里，这个rpc方法名首字母是小写，特别注意！
        return self.rpcclient.call("dump", onlyVisibleNode)


class CocosJsPoco(Poco):
    """docstring for CocosJsPoco"""

    def __init__(self, addr=DEFAULT_ADDR, device=None, **options):
        if not isinstance(addr, (tuple, list, six.string_types)):
            raise TypeError('Argument "addr" should be `tuple[2]`, `list[2]` or `string` only. Got {}'
                            .format(type(addr)))

        try:
            if isinstance(addr, (list, tuple)):
                ip, port = addr
            else:
                port = urlparse(addr).port
                if not port:
                    raise ValueError
                ip = urlparse(addr).hostname
        except ValueError:
            raise ValueError('Argument "addr" should be a tuple[2] or string format. e.g. '
                             '["localhost", 5003] or "ws://localhost:5003". Got {}'.format(repr(addr)))

        agent = CocosJsPocoAgent(port, device, ip=ip)
        if 'action_interval' not in options:
            options['action_interval'] = 0.5
        super(CocosJsPoco, self).__init__(agent, **options)



# -*- encoding=utf8 -*-
__author__ = "UWA"
__title__ = "UWA"
__desc__ = """
该脚本是UWA提供的用于自动化测试的工具包。
提供了较稳定的 apk 安装方法，与 UWA SDK 中的 GOT 模式进行交互的接口，与PocoSDK优化版的交互接口等。
"""

import json
from airtestProject.airtest.core.api import *
from airtestProject.poco.drivers.android.uiautomation import AndroidUiautomationPoco
from airtestProject.poco.utils.simplerpc.utils import sync_wrapper
try:
    import run_airtest_config
    import run_got_util
    UWAPPLTest = True
    log("--当前为UWAPPL调试--")
except:
    # log("--当前为本地调试--")
    UWAPPLTest = False

import time
import logging


# log 对象
_logger = logging.getLogger("airtest")
_logger.debug("[UWA] import UWA")


@sync_wrapper
def SetPruningEnabled(pocoHdl, enabled):
    """
    开启或关闭UWA优化后的PocoSDK提供的剪枝优化功能。
    """
    return pocoHdl.agent.c.call("SetPruningEnabled", enabled)

@sync_wrapper
def CollectWeakWhitelist(pocoHdl):
    """
    自动扫描场景物体，检测哪些节点需要保留。在开启剪枝优化功能的情况下，每次切换场景或场景的UI变化比较大时，需要调用一次。
    """
    return pocoHdl.agent.c.call("CollectWeakWhitelist")

@sync_wrapper
def SetBlackList(pocoHdl, *args):
    """
    设置Hierarchy信息节点的黑名单。默认被保留的节点，如果在黑名单里，则会被剔除掉。
    """
    return pocoHdl.agent.c.call("SetBlackList", *args)

@sync_wrapper
def SetWhiteList(pocoHdl, *args):
    """
    设置Hierarchy信息节点的白名单。默认被剔除掉的节点，如果在白名单里，则会被保留。白名单优先于黑名单。
    """
    return pocoHdl.agent.c.call("SetWhiteList", *args)


@sync_wrapper
def SetBlockedAttributes(pocoHdl, *args):
    """
    设置被屏蔽的属性。
    PocoSDK获取到的Hierarchy属性有：
    "name","type","visible","pos","size","scale", "anchorPoint", "zOrders",
    "clickable","text","components",  "texture","tag","_ilayer","layer", "_instanceId"

    推荐屏蔽掉的信息有："Tag","texture", "scale", "clickable", "components", "_ilayer", "layer", "zOrders"

	如果某个属性被屏蔽掉导致脚本运行异常，重新调用该函数设置屏蔽列表即可。
    """

    if UWAPPLTest == True:
        from poco.drivers.std.HierarchyTranslator import translator_agent
        translator_agent.SetBlockAttrs(*args)
    return pocoHdl.agent.c.call("SetBlockedAttributes", *args)


class GOT_Test():
    """
    GOT Online & GPM 自动化工具类
    """

    ENABLED = True

    @classmethod
    def Connect(cls, poco):
        """
        连接 GOT 进行自动化测试，使 GOT GUI 不可交互，只能通过脚本控制
        :param poco: Unity Poco 实例
        :Example:
            >>> GOT_Test.Connect(poco)
        """
        _logger.debug("[UWA] connect")
        if cls.ENABLED and not poco is None:
            return poco.agent.c.call("UWA.Connect")

    @classmethod
    def Start(cls, poco, mode="default", config= None, platform=True):
        """
        开启 GOT 的某个模式 或者 GPM，支持：overview/mono/resources/lua/gpu/gpm
        :param poco: Unity Poco 实例
        :param mode: 模式分为 overview/mono/resources/lua/gpu/gpm
        :param config: Dict 类型，包含各个模块的配置。如果为 None，则使用默认配置
            config = {
                "overview.mode": 0,                  Overview 模式具体测试模式，0 表示自定义模式，1 表示极简模式，2 表示CPU模式，3 表示内存模式
                "overview.engine_cpu_stack": True,   Overview 模式下，是否开启引擎、C#逻辑代码 CPU 调用堆栈统计
                "overview.lua_cpu_stack": True,      Overview 模式下，是否开启 Lua 逻辑代码 CPU 调用耗时堆栈统计。前置条件：engine_cpu_stack 为 True
                "overview.lua_mem_stack": True,      Overview 模式下，是否开启 Lua 逻辑代码内存占用堆栈统计。前置条件：lua 为 True
                "overview.time_line": True,          Overview 模式下，是否开启 Timeline 统计。前置条件：engine_cpu_stack 为 True
                "overview.stack_detail": 1,          Overview 模式下，堆栈统计的细节控制，0 表示默认细节，1 表示详细细节，2 表示全堆栈（不开放）。前置条件：engine_cpu_stack 为 True
                "overview.unity_api": True,          Overview 模式下，是否开启 Unity API 调用统计。前置条件：engine_cpu_stack 为 True
                "overview.lua": True,                Overview 模式下，是否开启 Lua 测试
                "overview.lua_dump_step": 0,         Overview 模式下，Lua 测试采样间隔，-1 表示不支持Dump，0 表示手动Dump，N自动以N为周期进行Lua Dump。前置条件：lua 为 True
                "overview.resources": True,          Overview 模式下，是否开启资源统计，0 表示完全关闭Resources获取（不可Dump），1 表示只获取资源总数、总内存（可Dump），2 表示获取资源详细信息（可Dump）
                "overview.unity_loading": True,      Overview 模式下，是否开启资源管理功能
                "mono.mono_dump_step": 0,            Mono 模式下，Mono 测试采样间隔，0 表示手动Dump，N自动以N为周期进行Mono Dump
                "resources.unity_loading": True,     Resources 模式下，是否开启资源管理功能
                "lua.lua_dump_step": 0,              Lua 模式下，Lua 测试采样间隔，0 表示手动Dump，N自动以N为周期进行Lua Dump
                "lua.lua_cpu_stack": True,           Lua 模式下，是否开启 Lua 逻辑代码 CPU 调用耗时堆栈统计，固定为 True
                "lua.lua_mem_stack": True,           Lua 模式下，是否开启 Lua 逻辑代码内存占用堆栈统计，固定为 True
                "gpu.texture_analysis": True,        GPU 模式下，是否开启纹理分析
                "gpu.mesh_analysis": True            GPU 模式下，是否开启网格分析
            }
        :param platform: 是否是Android平台
        :Example:
            >>> config={"resources.unity_loading":True}
            >>> GOT_Test.Start(poco, mode="resources", config=config)
            >>> GOT_Test.Start(poco, mode="gpm")
        """

        if UWAPPLTest:
            if run_airtest_config.got_type is not None:
                mode = run_airtest_config.got_type

        if mode == "assets":
            mode = "resources"

        _logger.debug("[UWA] Start " + mode)
        _logger.debug("[UWA] Start Config " + str(config))

        if cls.ENABLED and not poco is None:
            if config is None:
                log("UWA GOT Start")
                res = poco.agent.c.call("UWA.Start", mode)
            else:
                log("UWA GOT Start With Config")
                res = poco.agent.c.call("UWA.Start", mode, json.dumps(config))

            time.sleep(3)
            if platform:
                a_poco = AndroidUiautomationPoco()
                if a_poco("android:id/button1").exists():
                    a_poco("android:id/button1").click()
            return res
        return True

    @classmethod
    def Stop(cls, poco):
        """
        停止 GOT 模式
        :param poco: Unity Poco 实例
        :Example:
            >>> GOT_Test.Stop(poco)
        """
        _logger.debug("[UWA] Stop")
        log("UWA GOT Stop")  # show in the reports

        if cls.ENABLED and not poco is None:
            return poco.agent.c.call("UWA.Stop")
        return True

    @classmethod
    def StatPoco(cls, poco, stat):
        """
        是否在报告中保留 Poco 的额外耗时
        :param poco: Unity Poco 实例
        :param stat: True/False
        :Example:
            >>> GOT_Test.StatPoco(poco, True)
        """
        _logger.debug("[UWA] StatPoco : " + str(stat))
        log("UWA GOT StatPoco")
        if cls.ENABLED and not poco is None:
            return poco.agent.c.call("UWA.StatPoco", stat)

    @classmethod
    def Dump(cls, poco, dump):
        """
        手动 dump，采集更多信息
        :param poco: Unity Poco 实例
        :param dump: 可用dump类型：mono/lua/resources/overdraw，类型为String
        :Example:
            >>> GOT_Test.Dump(poco, "mono")
        """
        _logger.debug("[UWA] Dump : " + dump)
        log("UWA GOT Dump")
        if cls.ENABLED and not poco is None:
            return poco.agent.c.call("UWA.Dump", dump)

    @classmethod
    def Note(cls, poco, note):
        """
        添加备注信息，显示在 UWA官网 对应报告的备注栏。
        :param poco: Unity Poco 实例
        :param note: 备注信息
        :Example:
            >>> GOT_Test.Note(poco, "test")
        """
        _logger.debug("[UWA] Note : " + note)
        log("UWA GOT Note")
        if cls.ENABLED and not poco is None:
            return poco.agent.c.call("UWA.Note", note)

    @classmethod
    def Tag(cls, poco, tag):
        """
        标记测试区间
        :param poco: Unity Poco 实例
        :param tag: 区间标记
        :Example:
            >>> GOT_Test.Tag(poco, "test")
        """
        _logger.debug("[UWA] Tag " + tag)
        log("UWA GOT Tag")
        if cls.ENABLED and not poco is None:
            res = poco.agent.c.call("UWA.Tag", tag)
            return res

    @classmethod
    def Upload(cls, poco, timeLimitS=3600, uploadTimeLimitS=600):
        """
        上传测试数据到 UWA 官网，仅可在UWA Pipeline中使用。
        :param poco: Unity Poco 实例
        :param timeLimitS: 测试时长限制
        :param uploadTimeLimitS: 上传时长限制
        :Example:
            >>> GOT_Test.Upload(poco, 7200, 3000)
        """
        _logger.debug("[UWA] Upload " + str(timeLimitS) + " " + str(uploadTimeLimitS))
        log("UWA GOT Upload With UWA Pipeline")
        if UWAPPLTest:
            run_got_util.UploadWithRecordId(poco, timeLimitS, uploadTimeLimitS)

    @classmethod
    def LocalUpload(cls, poco, account, password, projectID, timeLimitS=3600, uploadTimeLimitS=600):
        """
        本地上传，需要配置好UWA的账号、密码、项目ID等信息。
        :param poco: Unity Poco 实例
        :param account: UWA 账号
        :param password: UWA 账号 密码
        :param projectID: 项目ID
        :param timeLimitS: 测试时长限制,默认3600秒
        :param uploadTimeLimitS: 上传时长限制,默认600秒
        :Example:
            >>> GOT_Test.LocalUpload(poco, "uwa", "111111", 111, 3000)
        """
        log("UWA GOT Upload With Local Test")
        if cls.ENABLED and not poco is None:
            poco.agent.c.call("UWA.StartUploadUserpwd", account, password, projectID, timeLimitS)

            upload_time = 0
            while upload_time < uploadTimeLimitS:
                upload_time += 5
                time.sleep(5)
                res, n = poco.agent.c.call("UWA.CheckUploadResult").wait()
                _logger.debug("[UWA] CheckUploadResult " + str(res))
                if len(res) == 3 and res[0]:
                    return res[1], res[2]

            raise Exception("UWA Upload Timeout")



"""
登录操作
"""
import random

from random import choice


from airtestProject.airtest.core.api import *
from airtestProject.commons.utils.logger import log
from airtestProject.factory.operateFactory import operate

#


reset_pos = [0, 0]

login_pos = 'btnLogin'  # 进入选择服务器界面
input_account_pos = "inputFieldAccountName"  # 输入账号
input_account_text_pos = "inputFieldAccountName-Text"  # 输入账号-文本
change_server_pos = "btnChangedServer"  # 选择服务器
enter_game_pos = "btnEnterGame"  # 进入选角界面
create_role_page = "CreateRoleView(Clone)"  # 创建角色界面
select_role_page = "SelectRoleView(Clone)"  # 选择角色界面
input_name_pos = "InputField"  # 输入创角名字
enter_pos = "Button_Enter"  # 进入游戏按钮
loading_page = "LoadingView(Clone)"  # 登陆界面
map_pos = "btnMinMap"  # 游戏内小地图按钮
tip_pos = "tipsLab"

pos_pos = "inputFieldAccountName"

all_role_pos = {"liuerlong": "0", "longnv": "1", "kuangzhanshi": "2", "tangsan": "3", "mahongjun": "4"}  # 角色按钮
gender_pos = {"male": 'Gender-Toggle', "female": 'Toggle (1)'}  # 性别按钮
account = ["luohaifeng", "luohaifeng02", "luohaifeng03", "test01", "test02", "test03", "test04", "test05", "test06",
           "test07", "test08", "test09"]  # 账号数据
all_server_pos = {"public": "0", "yuanyao": "1"}  # 服务器按钮


# for i in range(10):
#     if operate("poco").exists(login_pos):
#         log.step('进入登录界面成功')
#         break
#     else:
#         log.step('进入登录界面失败，等待进入')
#         operate("poco").sleep(10.0)


def r_user_name():
    # 生成一个包含所有字母和数字的字符集
    chinese_chars = ''.join(chr(i) for i in range(0x4e00, 0x9fa6))
    # 使用random.choice随机选择字符，生成指定长度的字符串
    random_string = ''.join(random.choice(chinese_chars) for _ in range(6))
    return random_string


class LoginPage:
    def __init__(self, adb, script_root, project=None, log_path=None):
        """

        :param project: 如果想采用命名代替文件夹路径的方法需要传入一个文件夹名让air生成对应字典。
        """
        self.adb = adb
        if project is not None:
            operate("air").set_dict(script_root, project)
        if log_path is not None:
            self.log_path = log_path
        else:
            self.log_path = None

    @log.wrap("进入登陆页面")
    def check_enter_login_view(self, pos, fun_name="air"):
        self.adb.shell("logcat -c")
        if operate(fun_name).wait_element_appear(pos):
            log.step("进入登录界面成功")
            return True
        else:
            log.step("进入登录界面失败")
            adb_log = self.adb.shell("logcat -d Unity:W *:S")
            log.log_adb_out(adb_log)
            return False

    @log.wrap('输入账号')
    def input_account(self, account_pos, fun_name="air"):
        user = choice(account)
        operate(fun_name).set_text(account_pos, ''.join(str(random.randint(1, 9)) for _ in range(6)))
        log.step(f'输入账号-{user}')

    @log.wrap('点击登录')
    def click_login(self, login_button, into_game_btn, fun_name="air"):
        self.adb.shell("logcat -c")
        """

        :param login_button: 登陆按钮控件
        :param into_game_btn: 下一个页面要出现的元素
        :param fun_name: air或poco
        :return:
        """
        start_time = time.time()
        while operate(fun_name).exists(into_game_btn) is False:
            operate(fun_name).click(login_button)
            if time.time() - start_time > 120:
                log.step("疑似点击登录失败")
                adb_log = self.adb.shell("logcat -d Unity:W *:S")
                log.log_adb_out(adb_log)
        operate(fun_name).sleep(1.0)

    @log.wrap('打开服务器列表')
    def open_server_list(self, change_server_pos, fun_name="air"):
        """

        :param change_server_pos:
        :param fun_name: air或poco
        :return:
        """
        for i in range(5):
            if operate(fun_name).exists(change_server_pos):
                operate(fun_name).click(change_server_pos)
                break
            else:
                log.error('找不到登录界面，尝试等待')
                operate(fun_name).sleep(1.0)
        operate(fun_name).sleep(1.0)

    @log.wrap('选择区服')
    def select_server(self, server_name, fun_name="air"):
        operate(fun_name).click(server_name, focus=reset_pos)
        operate(fun_name).sleep(1.0)

    @log.wrap('进入选角界面')
    def click_enter_game(self, enter_game_pos, fun_name="air"):
        """

        :param enter_game_pos: 进入游戏按钮
        :param fun_name: air或poco
        :return:
        """
        for i in range(3):
            if operate(fun_name).exists(enter_game_pos):
                operate(fun_name).click(enter_game_pos)
            else:
                return False
        operate(fun_name).sleep(2.0)

    def is_create_role(self, fun_name="air"):
        if operate(fun_name).exists(create_role_page):
            return True
        return False

    @log.wrap("选择角色")
    def select_role(self, role, fun_name="air"):
        operate(fun_name).click(role)

    @log.wrap("选择性别")
    def select_gender(self, gender, fun_name="air"):
        operate(fun_name).click(gender)

    @log.wrap('创建角色or选择角色')
    def create_role(self, create_role_page, select_role_page, role, gender, input_name_pos, enter_game_pos, create_pos,
                    next_page_pos, fun_name="air"):
        """

        :param two_page_list: 双页面等待，任何一个页面出现都会执行一下步
        :param role: 要创建的角色
        :param gender: 要创建的性别
        :param fun_name: air或poco
        :return:
        """
        operate(fun_name).wait_for_any([create_role_page, select_role_page])
        bool = operate(fun_name).exists(create_role_page)
        log.step(f'是否未创角-{bool}')
        # random_name = factory.random_word()
        if bool:
            operate(fun_name).click(role, focus=reset_pos)
            operate(fun_name).click(gender, focus=reset_pos, ocrPlus=True)
            operate(fun_name).set_text(input_name_pos, r_user_name())
            operate(fun_name).sleep(1.0)
            log.step(f'创建角色-test')
            for i in range(3):
                if operate(fun_name).exists(next_page_pos) is False:
                    operate(fun_name).click(create_pos)
                    log.step('点击进入游戏')
                else:
                    break
        else:
            for i in range(3):
                if operate(fun_name).exists(next_page_pos) is False:
                    operate(fun_name).click(enter_game_pos)
                    log.step('点击进入游戏')
                else:
                    break

    @log.wrap('检查进入游戏结果')
    def check_enter_success(self,tip_pos, fun_name="air"):
        """

        :param tip_pos: loding页面存在的元素
        :param fun_name: air或poco
        :return:
        """
        if operate(fun_name).wait_disappear_element(tip_pos) is True:
            log.step('进入游戏成功')
            return True
        else:
            log.step('进入游戏失败')
            adb_log = self.adb.shell("logcat -d Unity:W *:S")
            log.log_adb_out(adb_log)
            return False

    def odin_login(self):
        self.check_enter_login_view(login_pos, "poco")
        self.input_account(input_account_pos, "poco")
        self.click_login(login_pos, change_server_pos, "poco")
        self.open_server_list(change_server_pos, "poco")
        self.select_server(all_server_pos["public"], "poco")
        self.click_enter_game(enter_game_pos, "poco")
        self.create_role(create_role_page, select_role_page, create_role_page, all_role_pos['tangsan']
                         , gender_pos['male'], input_name_pos, enter_pos, enter_pos, tip_pos, "poco")
        # self.check_login_success()
        if self.check_enter_success(tip_pos, "poco"):
            return True

    def odin_login_air(self):
        self.check_enter_login_view("LoginBtn")
        self.input_account("AccountInputBox")
        self.click_login("LoginBtn", "ZoneChangeButton")
        self.open_server_list("ZoneChangeButton")
        self.select_server("Odin外网共公服")
        self.click_enter_game("EnterGame")
        self.create_role("CreateRole", "EnterGame (SelectCharacter)", "马红俊", "女",
                               "NameTxt", "EnterGame (SelectCharacter)", "CreateRole", "WaitingBar")
        if self.check_enter_success("WaitingBar"):
            return True


if __name__ == '__main__':
    # poco("inputFieldAccountName").set_text("fdshj")
    connect_device("Android:///e7970f92")
    # odin_login_air()



import os, sys
import subprocess, threading, time

from airtestProject.manager.LogManager import LogManager, catch_error

LIBASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(LIBASE_DIR)

log_manager = LogManager(logfile=None)


# 指定目录执行cmd指令，阻塞进程
@catch_error
def run_command(command, cwd="."):
    data = ({})
    try:
        p = subprocess.Popen(command, cwd=cwd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        std_out, std_err = p.communicate()
        ret_code = p.returncode

        data.update({"dir": cwd,
                     "cmd": command,
                     "code": ret_code,
                     "std_out": std_out,
                     "std_err": std_err,
                     })
        return data
    except Exception as e:
        import traceback
        log_manager.log_error(traceback.format_exc())
        return data


def _thread_run_command(command, log_file, cwd=".", ):
    try:
        p = subprocess.Popen(command, cwd=cwd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        while p.poll() is None:
            std_out = p.stdout.readline()
            if std_out is not None and len(std_out) > 0:
                with open(log_file, "ab") as file_obj:
                    file_obj.write(std_out)
    except Exception as e:
        import traceback
        log_manager.log_error(traceback.format_exc())


def random_string(length):
    import random, string
    if length > 0:
        max_once = 62
        if length > max_once:
            ret = "".join(random.sample(string.ascii_letters + string.digits, max_once)) \
                  + random_string(length - max_once)
        else:
            ret = "".join(random.sample(string.ascii_letters + string.digits, length))
        return ret
    else:
        return ""


# 指定目录执行cmd指令，异步执行不阻塞进程
# 执行结果输出到指定日志，返回值为日志路径
def run_command_async(command, cwd=".", log_name=None):
    crd = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))),
                       "log",
                       "tmp")

    if not os.path.isdir(crd):
        os.makedirs(crd)

    if log_name is None:
        log_name = "run_%s_%s.log" % (time.strftime('%Y%m%d%H%M%S', time.gmtime(time.time() + 60 * 60 * 8)),
                                      random_string(10).lower())

    log_file = os.path.join(crd, log_name)

    t = threading.Thread(target=_thread_run_command, args=(command, log_file, cwd))
    t.setDaemon(False)  # 设置为守护线程,主线程结束,同时全部子线程被终止

    t.start()

    return log_file



def adb_shell(cmd, device_id=None):
    if not device_id:
        run_cmd = f'adb shell {cmd}'
        result = subprocess.Popen(run_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()[
            0].decode("utf-8").strip()
        return result

    if type(device_id) != list:
        device_id = [device_id]
    for id in device_id:
        run_cmd = f'adb -s {id} shell {cmd}'
        result = subprocess.Popen(run_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()[
            0].decode("utf-8").strip()

    return True


if __name__ == "__main__":
    s = adb_shell("ps | filter findstr com.global.bcslg", "d1236ad8")
    print(111, s)
